# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest
from typing import Dict
from typing import Optional
from typing import Union
from typing import cast
from unittest import TestCase
from unittest.mock import patch

import numpy as np
import pandas as pd
from parameterized.parameterized import parameterized

from kats.compat import statsmodels
from kats.compat.pandas import assert_frame_equal
from kats.compat.pandas import assert_series_equal
from kats.consts import TimeSeriesData
from kats.data.utils import load_air_passengers
from kats.data.utils import load_data
from kats.models.theta import ThetaModel
from kats.models.theta import ThetaParams
from kats.tests.models.test_models_dummy_data import AIR_FCST_15_THETA_INCL_HIST_SM_11
from kats.tests.models.test_models_dummy_data import AIR_FCST_15_THETA_INCL_HIST_SM_12
from kats.tests.models.test_models_dummy_data import AIR_FCST_15_THETA_SM_11
from kats.tests.models.test_models_dummy_data import AIR_FCST_15_THETA_SM_12
from kats.tests.models.test_models_dummy_data import NONSEASONAL_INPUT
from kats.tests.models.test_models_dummy_data import (
    PEYTON_FCST_30_THETA_INCL_HIST_SM_11,
)
from kats.tests.models.test_models_dummy_data import (
    PEYTON_FCST_30_THETA_INCL_HIST_SM_12,
)
from kats.tests.models.test_models_dummy_data import PEYTON_FCST_30_THETA_SM_11
from kats.tests.models.test_models_dummy_data import PEYTON_FCST_30_THETA_SM_12

TEST_DATA: Dict[str, Dict[str, Union[ThetaParams, TimeSeriesData, pd.DataFrame]]] = {
    "short": {
        "ts": TimeSeriesData(
            pd.DataFrame(
                {
                    "time": [
                        pd.Timestamp("1961-01-01 00:00:00"),
                        pd.Timestamp("1961-02-01 00:00:00"),
                    ],
                    "y": [1.0, 2.0],
                }
            )
        ),
        "params": ThetaParams(m=2),
    },
    "constant": {
        "ts": TimeSeriesData(pd.DataFrame({"time": pd.date_range("1960-12-01", "1963-01-01", freq="m"), "y": 10.0})),
        "params": ThetaParams(m=2),
    },
    "nonseasonal": {
        "ts": TimeSeriesData(NONSEASONAL_INPUT),
        "params": ThetaParams(m=4),
    },
    "daily": {
        "ts": TimeSeriesData(load_data("peyton_manning.csv").set_axis(["time", "y"], axis=1)),
        "params": ThetaParams(),
        "params_negative": ThetaParams(m=-5),
    },
    "monthly": {
        "ts": load_air_passengers(),
        "params": ThetaParams(m=12),
    },
    "multivariate": {"ts": TimeSeriesData(load_data("multivariate_anomaly_simulated_data.csv"))},
}


class ThetaModelTest(TestCase):
    def test_params(self) -> None:
        # Test default value
        params = ThetaParams()
        params.validate_params()
        self.assertEqual(params.m, 1)

        params = ThetaParams(m=12)
        self.assertEqual(params.m, 12)

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator `parameter...
    @parameterized.expand(
        [
            [
                "monthly",
                TEST_DATA["monthly"]["ts"],
                TEST_DATA["monthly"]["params"],
                15,
                0.05,
                False,
                None,
                (AIR_FCST_15_THETA_SM_11 if statsmodels.version < "0.12" else AIR_FCST_15_THETA_SM_12),
            ],
            [
                "monthly, include history",
                TEST_DATA["monthly"]["ts"],
                TEST_DATA["monthly"]["params"],
                15,
                0.05,
                True,
                None,
                (
                    AIR_FCST_15_THETA_INCL_HIST_SM_11
                    if statsmodels.version < "0.12"
                    else AIR_FCST_15_THETA_INCL_HIST_SM_12
                ),
            ],
            [
                "daily",
                TEST_DATA["daily"]["ts"],
                TEST_DATA["daily"]["params"],
                30,
                0.05,
                False,
                None,
                (PEYTON_FCST_30_THETA_SM_11 if statsmodels.version < "0.12" else PEYTON_FCST_30_THETA_SM_12),
            ],
            [
                "daily, include history",
                TEST_DATA["daily"]["ts"],
                TEST_DATA["daily"]["params_negative"],
                30,
                0.05,
                True,
                None,
                (
                    PEYTON_FCST_30_THETA_INCL_HIST_SM_11
                    if statsmodels.version < "0.12"
                    else PEYTON_FCST_30_THETA_INCL_HIST_SM_12
                ),
            ],
        ]
    )
    def test_forecast(
        self,
        testcase_name: str,
        ts: TimeSeriesData,
        params: ThetaParams,
        steps: int,
        alpha: float,
        include_history: bool,
        freq: Optional[str],
        truth: pd.DataFrame,
    ) -> None:
        np.random.seed(0)
        m = ThetaModel(data=ts, params=params)
        m.fit()
        forecast_df = m.predict(steps=steps, alpha=alpha, include_history=include_history, freq=freq)
        assert_frame_equal(truth, forecast_df)

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator `parameter...
    @parameterized.expand(
        [
            [
                "m less than 1",
                TEST_DATA["daily"]["ts"],
                TEST_DATA["daily"]["params_negative"],
                False,
            ],
            [
                "data too short",
                TEST_DATA["short"]["ts"],
                TEST_DATA["short"]["params"],
                False,
            ],
            [
                "constant data",
                TEST_DATA["constant"]["ts"],
                TEST_DATA["constant"]["params"],
                False,
            ],
            [
                "seasonal",
                TEST_DATA["monthly"]["ts"],
                TEST_DATA["monthly"]["params"],
                True,
            ],
        ]
    )
    def test_check_seasonality(
        self,
        testcase_name: str,
        ts: TimeSeriesData,
        params: ThetaParams,
        is_seasonal: bool,
    ) -> None:
        m = ThetaModel(ts, params)
        m.check_seasonality()
        self.assertEqual(m.seasonal, is_seasonal)

    # pyre-fixme[56]: Pyre was not able to infer the type of the decorator `parameter...
    @parameterized.expand(
        [
            [
                "nonseasonal",
                False,
                TEST_DATA["nonseasonal"]["ts"],
                TEST_DATA["nonseasonal"]["params"],
                False,
                True,
            ],
            [
                "seasonal",
                True,
                TEST_DATA["monthly"]["ts"],
                TEST_DATA["monthly"]["params"],
                True,
                False,
            ],
        ]
    )
    def test_deseasonalize(
        self,
        testcase_name: str,
        seasonal: bool,
        ts: TimeSeriesData,
        params: ThetaParams,
        seasonality_removed: bool,
        decomp_is_none: bool,
    ) -> None:
        m = ThetaModel(ts, params)
        m.seasonal = seasonal
        deseas_data = m.deseasonalize()

        if seasonality_removed:
            self.assertFalse(ts.value.equals(deseas_data.value))
        else:
            assert_series_equal(cast(pd.Series, ts.value), cast(pd.Series, deseas_data.value))

        self.assertEqual(decomp_is_none, m.decomp is None)

    def test_multivar(self) -> None:
        # Theta model does not support multivariate time data
        self.assertRaises(
            ValueError,
            ThetaModel,
            TEST_DATA["multivariate"]["ts"],
            ThetaParams(),
        )

    def test_exec_plot(self) -> None:
        m = ThetaModel(
            cast(TimeSeriesData, TEST_DATA["daily"]["ts"]),
            cast(ThetaParams, TEST_DATA["daily"]["params"]),
        )
        m.fit()
        m.predict(steps=15, alpha=0.05)
        m.plot()

    def test_name(self) -> None:
        m = ThetaModel(
            cast(TimeSeriesData, TEST_DATA["daily"]["ts"]),
            cast(ThetaParams, TEST_DATA["daily"]["params"]),
        )
        self.assertEqual(m.__str__(), "Theta")

    def test_search_space(self) -> None:
        self.assertEqual(
            ThetaModel.get_parameter_search_space(),
            [
                {
                    "name": "m",
                    "type": "choice",
                    "values": list(range(1, 31)),
                    "value_type": "int",
                    "is_ordered": True,
                },
            ],
        )

    def test_others(self) -> None:
        m = ThetaModel(
            cast(TimeSeriesData, TEST_DATA["daily"]["ts"]),
            cast(ThetaParams, TEST_DATA["daily"]["params"]),
        )
        # fit must be called before predict
        self.assertRaises(ValueError, m.predict, 30)

        # seasonal data must be deseasonalized before fit
        with patch.object(m, "deseasonalize", (lambda self: self.data).__get__(m)), patch.object(
            m, "check_seasonality"
        ):
            m.seasonal = True
            m.decomp = None
            self.assertRaises(ValueError, m.fit)

        with patch(
            "kats.utils.decomposition.TimeSeriesDecomposition.decomposer",
            return_value={"seasonal": cast(TimeSeriesData, TEST_DATA["daily"]["ts"]) * 0},
        ):
            # Don't deseasonalize if any seasonal index = 0
            deseas_data = m.deseasonalize()
            expected = cast(pd.Series, cast(TimeSeriesData, TEST_DATA["daily"]["ts"]).value)
            assert_series_equal(expected, cast(pd.Series, deseas_data.value))


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""Ensemble Prediction Interval (EPI)

This is an Ensemble method to estimate the prediction interval for any forecasting models
The high level idea is to estimate the empirical error matrix (S) from a specific
forecasting model, and then calculate mean (m) and standard deviation (std) from S.
When doing forecasting, adjust the original fcst by a random sample generated from N(m, std).
Do this procedure for m times (ensemble), based on which, generate fcst/fcst_upper/fcst_lower.
"""

import logging
from multiprocessing import cpu_count
from multiprocessing.pool import Pool
from typing import Optional
from typing import Tuple
from typing import Type

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy

from kats.consts import Params
from kats.consts import TimeSeriesData
from kats.models.model import Model

_LOGGER: logging.Logger = logging.getLogger()


def mean_confidence_interval(
    data: np.ndarray,
    confidence: float = 0.8,
) -> Tuple[float, float, float]:
    data = 1.0 * np.array(data)
    dof = len(data)
    # calculate mean and std
    m, se = np.mean(data), np.std(data)
    h = se * scipy.stats.t.ppf((1 + confidence) / 2.0, dof - 1)
    return m, m - h, m + h


class ensemble_predict_interval:
    """
        Class for Ensemble Prediction Interval.

        The steps are listed as follows:
        1. Split the given ts into k (k = n_block + 1) parts, p_1 to p_k. Each part has size as block_size.
        2. For given model and params, train the model with data p_i, and predict next block_size step,
            and calculate the bias vector by comparing it to p_(i+1).
        3. We get an error matrix S with shape n_block * block_size from step 2, and then calculate mean (m)
            and standard deviation (std) from S.
        4. When doing forecasting, adjust the original fcst by a random sample generated from N(m, std).
        5. Repeat step 4 for m times (ensemble_size), based on which, generate fcst/fcst_upper/fcst_lower.

    Attributes:
        model: forecasting model
        model_params: forecasting model parameters
        ts: the time series data in `TimeSeriesData` format
        block_size: the size of blocks
        n_block: number of block we want to split the ts into
        ensemble_size: size of ensemble

    """

    def __init__(
        self,
        model: Type[Model[Params]],
        model_params: Params,
        ts: TimeSeriesData,
        block_size: Optional[int] = None,
        n_block: Optional[int] = None,
        ensemble_size: int = 10,
        multiprocessing: bool = False,
    ) -> None:
        # detection model
        self.model: Type[Model[Params]] = model
        # detection model params
        self.params: Params = model_params

        # check if the given block_size and n_block are reasonable
        if block_size is None and n_block is None:
            raise ValueError("Please provide an initial value for either block_size or n_block.")
        elif block_size is None:
            assert n_block is not None
            if n_block < 5:
                raise ValueError(f"The given n_block is {n_block}. Please provide a larger block_size.")
            self.n_block: int = n_block
            block_size = len(ts) // (self.n_block + 1)
            if block_size < 10:
                raise ValueError(f"Block_size is {block_size}. Please provide a longer TS or a smaller n_block.")
            self.block_size: int = block_size
        elif n_block is None:
            assert block_size is not None
            if block_size < 10:
                raise ValueError(f"The given block_size is {block_size}. Please provide a larger n_block.")
            self.block_size: int = block_size
            n_block = len(ts) // self.block_size - 1
            if n_block < 5:
                raise ValueError(f"n_block is {n_block}. Please provide a longer TS or a smaller block_size.")
            self.n_block: int = n_block
        else:
            if block_size < 10:
                raise ValueError(f"The given block_size is {block_size}. Please provide a larger n_block.")
            if n_block < 5:
                raise ValueError(f"The given n_block is {n_block}. Please provide a larger block_size.")
            if len(ts) < (n_block + 1) * block_size:
                raise ValueError(
                    f"The given TS has length {len(ts)}, which is samller than (n_block + 1) * block_size. Please provide a longer TS."
                )
            self.block_size: int = block_size
            self.n_block: int = n_block

        # time series data
        self.ts: TimeSeriesData = ts[-self.block_size * (self.n_block + 1) :]

        # infer freqency, data granularity
        self.freq: str = str(int(self.ts.infer_freq_robust().total_seconds())) + "s"

        # check ensemble size
        if ensemble_size < 4:
            raise ValueError(f"The given ensemble_size is {ensemble_size}. Please provide a larger n_block.")
        self.ensemble_size: int = ensemble_size

        # initial error matrix
        self.error_matrix: np.ndarray = np.empty([self.n_block, self.block_size])
        self.error_matrix_flag: bool = False

        # initial fcst results
        self.ensemble_fcst: np.ndarray = np.empty([1, 1])
        self.projection_flag: bool = False

        self.multiprocessing: bool = multiprocessing

    def _get_error_matrix_single(self, idx: int) -> Tuple[int, np.ndarray]:
        """
        Calculate error vector for each block
        """
        train_ts = TimeSeriesData(
            pd.DataFrame(
                {
                    "time": self.ts[idx * self.block_size : (idx + 1) * self.block_size].time.to_list(),
                    "value": self.ts[idx * self.block_size : (idx + 1) * self.block_size].value.to_list(),
                }
            )
        )
        m = self.model(train_ts, self.params)
        m.fit()
        pred = m.predict(steps=self.block_size, freq=self.freq)
        assert pred is not None
        fcst = pred["fcst"]

        sigma = np.asarray(
            self.ts[(idx + 1) * self.block_size : (idx + 2) * self.block_size].value.to_list()
        ) - np.asarray(fcst)

        return idx, sigma

    def _get_error_matrix(self) -> None:
        """
        Get an error vector for each block and combine them as a matrix
        """
        if self.error_matrix_flag:
            return

        if self.multiprocessing:
            num_process = max(min(self.n_block, (cpu_count() - 1) // 2), 1)
            with Pool(processes=num_process) as pool:
                records = pool.map(self._get_error_matrix_single, list(range(self.n_block)))
                pool.close()

            # combine results
            for idx, single_sigma in records:
                self.error_matrix[idx, :] = single_sigma
        else:
            for i in range(self.n_block):
                _, single_sigma = self._get_error_matrix_single(i)
                self.error_matrix[i, :] = single_sigma

        self.error_matrix_flag = True

    def _projection_single(self, idx: int, step: int = 30, rolling_based: bool = False) -> Tuple[int, np.ndarray]:
        """
        Get forecasting for future steps for one chain.
        """
        future_block = step // self.block_size + int(step % self.block_size > 0)

        onechain_fcst = np.zeros(future_block * self.block_size)
        if not rolling_based:
            training_data = TimeSeriesData(
                pd.DataFrame(
                    {
                        "time": self.ts[-self.block_size :].time.to_list(),
                        "value": self.ts[-self.block_size :].value.to_list(),
                    }
                )
            )
        else:
            training_data = TimeSeriesData(
                pd.DataFrame(
                    {
                        "time": self.ts[:].time.to_list(),
                        "value": self.ts[:].value.to_list(),
                    }
                )
            )
        for i in range(future_block):
            m = self.model(training_data, self.params)
            m.fit()
            pred = m.predict(steps=self.block_size, freq=self.freq)
            assert pred is not None
            fcst = np.asarray(pred["fcst"])

            fcst[:] += np.random.multivariate_normal(self.error_matrix.mean(0), np.cov(self.error_matrix.T), 1)[0]

            onechain_fcst[i * self.block_size : (i + 1) * self.block_size] = fcst.copy()

            # refresh training_data
            if not rolling_based:
                training_data = TimeSeriesData(pd.DataFrame({"time": pred["time"].tolist(), "value": list(fcst)}))
            else:
                training_data.extend(
                    TimeSeriesData(pd.DataFrame({"time": pred["time"].tolist(), "value": list(fcst)}))
                )

        return idx, onechain_fcst[:step]

    def _projection(self, step: int = 30, rolling_based: bool = False) -> None:
        """
        Get forecasting for future steps.
        """
        if not self.error_matrix_flag:
            self._get_error_matrix()

        ensemble_fcst = np.zeros([self.ensemble_size, step])
        if self.multiprocessing:
            num_process = max(min(self.ensemble_size, (cpu_count() - 1) // 2), 1)
            ipt_list = [[i, step, rolling_based] for i in range(self.ensemble_size)]
            with Pool(processes=num_process) as pool:
                records = pool.starmap(self._projection_single, ipt_list)
                pool.close()

            # combine results
            for idx, single_fcst in records:
                ensemble_fcst[idx, :] = single_fcst
        else:
            for i in range(self.ensemble_size):
                _, fcst = self._projection_single(i, step, rolling_based)
                ensemble_fcst[i, :] = fcst

        self.ensemble_fcst = ensemble_fcst

    def get_projection(
        self,
        step: int = 30,
        rolling_based: bool = False,
        confidence_level: float = 0.8,
    ) -> pd.DataFrame:
        """
        Get forecasting for future steps.
        """
        self._projection(step=step, rolling_based=rolling_based)
        mi, low, up = np.zeros(step), np.zeros(step), np.zeros(step)

        for i in range(step):
            mi[i], low[i], up[i] = mean_confidence_interval(self.ensemble_fcst[:, i], confidence_level)

        res = pd.DataFrame({"fcst": list(mi), "fcst_lower": list(low), "fcst_upper": list(up)})

        self.projection_flag = True
        return res

    def get_fcst_band_with_level(
        self,
        confidence_level: float,
    ) -> pd.DataFrame:
        """
        Get forecasting band for future steps with other confidence level.
        """
        if not self.projection_flag:
            raise ValueError("Please run get_projection() first.")

        step = self.ensemble_fcst.shape[1]
        mi, low, up = np.zeros(step), np.zeros(step), np.zeros(step)

        for i in range(step):
            mi[i], low[i], up[i] = mean_confidence_interval(self.ensemble_fcst[:, i], confidence_level)

        res = pd.DataFrame({"fcst": list(mi), "fcst_lower": list(low), "fcst_upper": list(up)})

        return res

    def pi_comparison_plot(
        self,
        test_ts: Optional[TimeSeriesData] = None,
        confidence_level: float = 0.8,
        figure_size: Tuple[int, int] = (10, 5),
        test_data_only: bool = False,
    ) -> None:
        if test_data_only and not test_ts:
            raise ValueError("Please provide test_ts.")

        T = self.ensemble_fcst.shape[1]
        if not self.projection_flag:
            raise ValueError("Please train and predit the model first.")

        if test_data_only:
            hist_data = []
            hist_data_len = 0
        else:
            hist_data = self.ts.value
            hist_data_len = len(hist_data)

        mi, low, up = np.zeros(T), np.zeros(T), np.zeros(T)

        for i in range(T):
            mi[i], low[i], up[i] = mean_confidence_interval(self.ensemble_fcst[:, i], confidence_level)

        fig, ax = plt.subplots()

        ax.plot(
            range(hist_data_len, T + hist_data_len),
            mi,
            lw=1,
            color="g",
            alpha=1,
            label="Predicted value from EPI",
        )
        ax.fill_between(
            range(hist_data_len, T + hist_data_len),
            low,
            up,
            color="g",
            alpha=0.4,
            label="Prediction interval band from EPI",
        )

        if hist_data_len > 0:
            ax.plot(
                range(0, hist_data_len),
                hist_data[:hist_data_len],
                lw=1,
                color="blue",
                alpha=1,
                label="Historical value",
            )
        if test_ts is not None:
            ax.plot(
                range(hist_data_len, min(len(test_ts), T) + hist_data_len),
                test_ts.value[: min(len(test_ts), T)],
                lw=1,
                color="r",
                alpha=1,
                label="Test value",
            )

        ax.set_xlabel("Time")
        ax.set_ylabel("Value")

        plt.legend(loc="upper left")
        fig.set_size_inches(*figure_size)
        plt.show()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


import logging
from typing import Any
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy.stats import chi2
from sklearn.covariance import MinCovDet

from kats.consts import TimeSeriesChangePoint
from kats.consts import TimeSeriesData
from kats.detectors.detector import Detector

"""A module for detecting abnormal in hourly ratio.

This module contains the class :class:`HourlyRatioDetector`, which detects the abnormal intra-day hourly ratio patterns.
"""


class HourlyRatioDetector(Detector):
    """Hourly Ratio Anormaly detector.

    This detector detects the abnormal intra-day hourly ratio patterns. This detector takes TimeSeriesDataas input and returns a list of TimeSeriesChangePoint representing the abnormal dates.
    The detection algorithm assumes that the hourly ratio of each day should follow a multivariate normal distribution, and we utilize Mahalanobis distance tests to detect abnormal days.
    This class provides detector and plot.

    Attributes:
        data: A :class:`kats.consts.TimeSeriesData` object representing the data to be examed, which should be of hour-level granularity.
        freq: Optional; A string or a `pandas.Timedelta` object representing the data frequency (following the naming conventions of pandas). Can be 'H' (hourly frequency), 'T' minutely frequency, 'S' secondly frequency or any other frequency finer than hourly frequency.
            Default is None, in which case the frequency will be infered by infer_freq_robust.
        aggregate: Optional; A string representing the aggregation method for aggregating data to hourly level data. Can be 'min', 'max', 'sum', 'mean' or None. Default is None, which means no aggregation.

    Sample Usage:
        >>> hr = HourlyRatioDetector(data)
        >>> anomlies=hr.detector()
        >>> hr = hr.plot(weekday = 3) # Plot the anomalies of weekday 3
    """

    def __init__(
        self,
        data: TimeSeriesData,
        freq: Union[str, pd.Timedelta, None] = None,
        aggregate: Optional[str] = None,
    ) -> None:
        super(HourlyRatioDetector, self).__init__(data=data)
        if len(data) == 0:
            msg = "Input data is empty."
            logging.error(msg)
            raise ValueError(msg)

        if not self.data.is_univariate():
            msg = "Only support univariate time series, but get {}.".format(type(self.data.value))
            logging.error(msg)
            raise ValueError(msg)
        self._ratiodf: Optional[pd.DataFrame] = None
        self.incomplete_dates: Optional[List[TimeSeriesChangePoint]] = None
        self.anomaly_dates: Optional[List[TimeSeriesChangePoint]] = None
        self.freq = freq
        self.aggregate = aggregate
        self._valid_frequency()

    def _valid_frequency(self) -> None:
        """Valid frequency of the input timeseries.

        If freq is given by user, then use the freq defined by user, otherwise we use ts.infer_freq_robust() to infer data frequencey.
        Data freq should be at least hourly level. For data with the granularity finner than hourly level, aggregate function should be given using aggregate attribuate.
        Now only support aggregation functions: min, max, sum, mean.
        """

        lower_granularity = ["T", "S", "L", "U", "N"]
        if self.freq is None:
            self.freq = self.data.infer_freq_robust()
        if self.freq == "H" or (isinstance(self.freq, pd.Timedelta) and self.freq.value == 3600000000000):
            msg = "Input data is hourly data."
            logging.info(msg)
            return
        if isinstance(self.freq, str):
            for level in lower_granularity:
                if isinstance(self.freq, str) and level in self.freq:
                    msg = "Input data granularity is {} and we can continue processing using aggregation function.".format(
                        self.freq
                    )
                    logging.info(msg)
        elif isinstance(self.freq, pd.Timedelta) and self.freq.value < 3600000000000:
            pass
        else:
            msg = "Time series should be of hourly or finer granularity."
            logging.error(msg)
            raise ValueError(msg)

        if self.aggregate is None:
            msg = "Aggregation method is missing."
            logging.error(msg)
            raise ValueError(msg)
        elif self.aggregate in ["min", "max", "sum", "mean"]:
            msg = "Aggregation method is {}.".format(self.aggregate)
            logging.info(msg)
            return
        else:
            msg = "Aggregation methd {} is not implemented.".format(self.aggregate)
            logging.error(msg)
            raise ValueError(msg)

    def _preprocess(self) -> None:
        """preprocess input data.

        There are two steps for preprocess: 1) filter out dates with incomplete data, aggregate data to hourly level if necessary; and 2) calculate hourly ratio.
        """

        if self._ratiodf is None:
            df = self.data.to_dataframe()
            df.dropna(inplace=True)
            df.sort_values("time", inplace=True)
            df["date"] = df["time"].dt.date
            df["hour"] = df["time"].dt.hour
            df["weekday"] = df["time"].dt.weekday
            # aggregate the data to hourly level.
            if self.freq != "H" and self.aggregate is not None:
                df = df.groupby(["date", "hour", "weekday"])["value"].agg(self.aggregate).reset_index()
                msg = "Successfully aggregate data to hourly level using {}".format(self.aggregate)
                logging.info(msg)
            df["counts"] = df.groupby("date")["hour"].transform("count")
            # filter out dates with less than 24 observations
            incomplete_dates = df["date"][df["counts"] < 24].unique()
            self.incomplete_dates = [TimeSeriesChangePoint(t, t, 1.0) for t in incomplete_dates]
            df = df[df["counts"] == 24]
            if len(df) == 0:
                msg = "Data should have hour-level granularity."
                logging.error(msg)
                raise ValueError(msg)
            df["hourly_mean"] = df.groupby("date")["value"].transform("sum")
            df["hourly_ratio"] = df["value"] / df["hourly_mean"]
            self._ratiodf = df
        return

    def _mahalanobis_test(
        self,
        obs: np.ndarray,
        median: np.ndarray,
        cov: np.ndarray,
        alpha: float = 0.01,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """mahalanobis test function.

        Args:
            obs: A :class:`numpy.ndarray` object storing the data to be tested.
            median: A :class:`numpy.ndarray` object storing the medians used to build centeralize test data.
            cov: A :class:`numpy.ndarray` object representing the covariance matrix.
            alpha: A float representing the significance level for the Mahalanobis test. We take the instance with pvalue<=alpha as an abnormal point.

        Returns:
            lab: A :class:`numpy.ndarray` object of booleans representing whether the corresponding instance is abnormal or not.
            pvalue: A :class:`numpy.ndarray` object storing the pvalues of tests of each instance.
        """

        diff = obs - median
        scores = np.sum(diff * np.linalg.solve(cov, diff.T).T, axis=1)
        pvalue = 1.0 - chi2(df=diff.shape[1]).cdf(scores)
        lab = pvalue <= alpha
        return (lab, pvalue)

    # pyre-fixme[14]: `detector` overrides method defined in `Detector` inconsistently.
    def detector(self, support_fraction: float = 0.9) -> Sequence[TimeSeriesChangePoint]:
        """Run detection algorithm.

        Args:
            support_fraction: Optional; A float representing the support_fraction for MinCovDet class from scikit-learn. Default is 0.9.
                              See https://scikit-learn.org/stable/modules/generated/sklearn.covariance.MinCovDet.html for more details.

        Returns:
            A list of TimeSeriesChangePoint representing the anormal dates.
        """

        if self._ratiodf is None:
            self._preprocess()
        anomaly = []
        pvalues = []
        for w in range(7):
            if self._ratiodf is None:
                msg = "self._ratiodf should not be none after running _preprocess"
                logging.error(msg)
                raise ValueError(msg)
            obs = self._ratiodf[self._ratiodf["weekday"] == w]["hourly_ratio"].values.reshape(-1, 24)
            if self._ratiodf is None:
                msg = "self._ratiodf should not be none after running _preprocess"
                logging.error(msg)
                raise ValueError(msg)
            dates = np.unique(self._ratiodf[self._ratiodf["weekday"] == w]["date"].values)
            # we omit the last dimension due to linearity constrant
            median = np.median(obs, axis=0)
            median = (median / np.sum(median) * 24)[:-1]
            cov = MinCovDet(assume_centered=True, support_fraction=support_fraction).fit(obs[:, :-1] - median)
            lab, p = self._mahalanobis_test(obs[:, :-1], median, cov.covariance_)
            anomaly.extend(list(dates[lab]))
            pvalues.extend(p[lab])
        anomaly = [TimeSeriesChangePoint(anomaly[i], anomaly[i], 1.0 - pvalues[i]) for i in range(len(anomaly))]
        self.anomaly_dates = anomaly
        return anomaly

    def plot(self, weekday: int = 0, **kwargs: Any) -> plt.Axes:
        """plot the detection results.

        Args:
            weekday: Optional; An integer representing the weekday label, which should be in [0,6]. Default is 0.

        Returns:
            The matplotlib Axes.
        """
        if self.anomaly_dates is None:
            msg = "Please run detector method first."
            logging.error(msg)
            raise ValueError(msg)
        anomaly_dates = [t.start_time for t in self.anomaly_dates]
        anomaly_dates = set(anomaly_dates)
        if self._ratiodf is None:
            msg = "Please run detector method first."
            logging.error(msg)
            raise ValueError(msg)
        obs = self._ratiodf[self._ratiodf["weekday"] == weekday]["hourly_ratio"].values.reshape(-1, 24)
        if self._ratiodf is None:
            msg = "Please run detector method first."
            logging.error(msg)
            raise ValueError(msg)
        dates = np.unique(self._ratiodf[self._ratiodf["weekday"] == weekday]["date"].values)
        labs = [(t in anomaly_dates) for t in dates]
        logging.info("# of anomaly dates: {}".format(np.sum(labs)))
        _, ax = plt.subplots()
        for i in range(len(obs)):
            if not labs[i]:
                ax.plot(obs[i], "--", color="silver", alpha=0.5)
            else:
                ax.plot(obs[i], "--", color="navy", label=str(dates[i]))
        ax.set_title("Hourly Ratio Plot for Weeday {weekday}")
        ax.legend()
        return ax



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
This file defines the data-partition classes for Kats.

Kats supports multiple types of data-partition classes, including:
    - :class: `SimpleDataPartition` (basic train & test data partition).
"""

import logging
from abc import ABC
from abc import abstractmethod
from datetime import datetime
from multiprocessing import cpu_count
from multiprocessing.dummy import Pool
from typing import Any
from typing import Dict
from typing import List
from typing import NamedTuple
from typing import Optional
from typing import Tuple
from typing import Union
from typing import cast

import numpy as np
import pandas as pd

from kats.consts import TimeSeriesData

Timestamp = Union[str, pd.Timestamp, datetime]
DataPartition = Union[TimeSeriesData, List[TimeSeriesData], Dict[Union[str, int], TimeSeriesData]]


class TrainTestData(NamedTuple):
    train: DataPartition
    test: DataPartition


def _raise_error(info: str) -> None:
    logging.error(info)
    raise ValueError(info)


def _get_absolute_size(size: int, frac: float) -> int:
    return int(size * frac)


class DataPartitionBase(ABC):
    """
    This class defines the parent functions for various data partition classes.

    Attributes:
        multi: A boolean indicating whether to use multiprocessing or not.
        max_core: Optional; An integer representing the number of cores for multiprocessing.
                  Default is None, which sets `max_core = max((total_cores - 1) // 2, 1)`.

    Raises:
        ValueError: An invalide data type is passed.
    """

    def __init__(self, multi: bool, max_core: Optional[int] = None, **kwargs: Any) -> None:
        self.multi = multi
        if multi:  # using multi-processing
            total_cores = cpu_count()
            if max_core is None:
                self.max_core: int = max((total_cores - 1) // 2, 1)
            elif isinstance(max_core, int) and max_core > 0 and max_core < total_cores:
                self.max_core = max_core
            elif isinstance(max_core, int) and max_core > total_cores:
                logging.warn(
                    f"`max_core` is larger than maximum available cores ({total_cores}) and sets `max_core = {total_cores}`."
                )
            else:
                msg = f"max_core should be a positive integer in [1, {total_cores}] but receives {max_core}."
                logging.error(msg)
                raise ValueError(msg)
        else:
            self.max_core = 1
        self.split_num: int = -1

    @abstractmethod
    def _single_train_test_split(
        self,
        tag: Union[int, str],
        data: TimeSeriesData,
    ) -> Tuple[Union[int, str], List[TrainTestData]]:
        raise NotImplementedError()

    def _data_check(
        self,
        data: Union[
            List[TimeSeriesData],
            Dict[Union[int, str], TimeSeriesData],
            TimeSeriesData,
        ],
    ) -> List[Union[int, str]]:
        keys = []
        if isinstance(data, TimeSeriesData):
            keys = [0]
            data = [data]
        # input is a dictionary
        elif isinstance(data, dict):
            keys = list(data.keys())
        elif isinstance(data, list):
            keys = list(range(len(data)))
        else:
            info = (
                "The input data should be a single (or a list or a dictionary of)"
                f"`pd.DataFrame`, `np.ndarray`, or `TimeSeriesData` object, but receives {type(data)}"
            )
            _raise_error(info)
        for k in keys:
            # pyre-fixme
            if not isinstance(data[k], TimeSeriesData):
                info = f"The input data should be a single (or a list or a dictionary of) `TimeSeriesData` object, but receives {type(data[k])}."
                _raise_error(info)
        # pyre-fixme
        return keys

    def split(self, data: DataPartition) -> List[TrainTestData]:
        """Splits the inputs data.

        Args:
            data: A single or a list or a dictionary of `TimeSeriesData` or `pd.DataFrame` or `np.ndarray` objects.

        Returns:
            A List of tuples and each tuple contrains two components, and the first component representing the training time series and the second component representing the test time series.
            More specifically, the returned data type is `List[tuple[train_ts, test_ts]]`.
        """

        # check data type
        keys = self._data_check(data)
        n = len(keys)
        # Only one time series
        if (not isinstance(data, list)) and (not isinstance(data, dict)):
            _, res = self._single_train_test_split(0, data)
            return res
        # Multiple time series
        if self.multi:
            # multi-processing
            # pyre-fixme
            params = [(k, data[k]) for k in keys]
            pool = Pool(self.max_core)
            raw_res = pool.starmap(self._single_train_test_split, params)
            pool.close()
            pool.join()
        else:
            # pyre-fixme
            raw_res = [self._single_train_test_split(k, data[k]) for k in keys]

        res = []
        return_list = isinstance(data, list)
        for i in range(self.split_num):
            tmp_train = {raw_res[k][0]: (raw_res[k][1][i]).train for k in range(n)}
            tmp_test = {raw_res[k][0]: (raw_res[k][1][i]).test for k in range(n)}
            if return_list:
                tmp_train = [tmp_train[k] for k in keys]
                tmp_test = [tmp_test[k] for k in keys]
            res.append(TrainTestData(train=tmp_train, test=tmp_test))
        return res


class SimpleDataPartition(DataPartitionBase):
    """SimpleDataPartition Class provides method for spliting time series into train/test sets given train/test data fraction or data number.

    Attributes:
        train_frac: Optional; A float in (0,1) representing the fraction of data for training, i.e., the first `train_frac*len(data)` data will be training data. Default is None.
        test_frac: Optional; A float in (0,1) representing the fraction of data for testing, i.e., the next `test_frac*len(data)` data after training data will be testing data. Default is None.
        train_size: Optional; An integer represening the number of data for training, i.e., the first `train_size` data will be training data. Default is None.
        test_size: Optional; An integer represening the number of data for testing, i.e., the next `test_size` data after training data will be testing data. Default is None.
        multi: Optional; A boolean representing whether using multi-processing or not. Default is True.
        max_core: Optional; An integer representing the number of cores for multiprocessing. Default is None, which sets `max_core = max((total_cores - 1) // 2, 1)`.

    Sample Usage:
        >>> sdp = SimpleDataPartition(train_frac = 0.8, test_frac = 0.2)
        >>> res = sdp.split(data)
    """

    def __init__(
        self,
        train_frac: Optional[float] = None,
        test_frac: Optional[float] = None,
        train_size: Optional[int] = None,
        test_size: Optional[int] = None,
        multi: bool = True,
        max_core: Optional[int] = None,
    ) -> None:

        super(SimpleDataPartition, self).__init__(multi=multi, max_core=max_core)
        self.split_num = 1
        if (train_frac is None) and (train_size is None):
            info = "Either `train_frac` or `train_size` should not be None."
            _raise_error(info)

        self._param_check(train_frac, train_size, "train_frac", "train_size")
        self._param_check(test_frac, test_size, "test_frac", "test_size")

        self.train_frac = train_frac
        self.train_size = train_size
        self.test_frac = test_frac
        self.test_size = test_size

    def _param_check(self, frac: Optional[float], size: Optional[int], frac_name: str, size_name: str) -> None:
        if (frac is not None) and (size is not None):
            info = f"Please specify either `{frac_name}` or `{size_name}`, not both."
            _raise_error(info)
        if (frac is not None) and (frac <= 0 or frac >= 1):
            info = f"`{frac_name}` should be a float in (0, 1), but receives {frac}."
        if (size is not None) and (size <= 0):
            info = f"`{size_name}` should be a positive integer, but receives {size}."
        return

    def _single_train_test_split(
        self,
        tag: Union[str, int],
        data: TimeSeriesData,
    ) -> Tuple[Union[int, str], List[TrainTestData]]:
        size = len(data)
        train_size = _get_absolute_size(size, self.train_frac) if self.train_frac else self.train_size
        train_size = cast(int, train_size)
        if self.test_frac:
            end_idx = _get_absolute_size(size, self.test_frac) + train_size
        elif self.test_size:
            end_idx = self.test_size + train_size
        else:
            end_idx = size
        return tag, [TrainTestData(train=data[:train_size], test=data[train_size:end_idx])]


class SimpleTimestampDataPartition(DataPartitionBase):
    """SimpleDataPartition Class provides method for spliting time series into train/test sets given timestamps.

    Attributes:
        train_end: A timestamp (i.e., string, `pd.Timestamp` object or `datetime.datetime` object) representing the end timestamp (inclusive) of training data.
        test_start: A timestamp (i.e., string, `pd.Timestamp` object or `datetime.datetime` object) representing the start timestamp (inclusive) of test data.
        train_start: Optional; A timestamp (i.e., string, `pd.Timestamp` object or `datetime.datetime` object) representing the start timestamp (inclusive) of training data. Default is None, which takes `pd.Timestamp.min`.
        test_end: Optional; A timestamp (i.e., string, `pd.Timestamp` object or `datetime.datetime` object) representing the end timestamp (inclusive) of test data. Default is None, which takes `pd.Timestamp.max`.
        multi: Optional; A boolean representing whether using multi-processing or not. Default is True.
        max_core: Optional; An integer representing the number of cores for multiprocessing. Default is None, which sets `max_core = max((total_cores - 1) // 2, 1)`.

    Sample Usage:
        >>> stdp = SimpleTimestampDataPartition(train_frac = 0.8, test_frac = 0.2)
        >>> res = stdp.split(data)
    """

    def __init__(
        self,
        train_end: Timestamp,
        test_start: Timestamp,
        train_start: Optional[Timestamp] = None,
        test_end: Optional[Timestamp] = None,
        multi: bool = True,
        max_core: Optional[int] = None,
    ) -> None:
        super(SimpleTimestampDataPartition, self).__init__(multi=multi, max_core=max_core)
        self.split_num = 1
        train_start, train_end, test_start, test_end = self._timestamp_check(
            train_start, train_end, test_start, test_end
        )

        self.train_start: pd.Timestamp = train_start
        self.train_end: pd.Timestamp = train_end
        self.test_start: pd.Timestamp = test_start
        self.test_end: pd.Timestamp = test_end

    def _timestamp_check(
        self,
        train_start: Optional[Timestamp],
        train_end: Timestamp,
        test_start: Timestamp,
        test_end: Optional[Timestamp],
    ) -> Tuple[pd.Timestamp, pd.Timestamp, pd.Timestamp, pd.Timestamp]:
        train_end = pd.Timestamp(train_end)
        test_start = pd.Timestamp(test_start)

        train_start = pd.Timestamp.min if train_start is None else pd.Timestamp(train_start)
        test_end = pd.Timestamp.max if test_end is None else pd.Timestamp(test_end)

        if train_end <= train_start:
            msg = "`train_end` should be greater than `train_start`."
            _raise_error(msg)
        if test_end <= test_start:
            msg = "`test_end` should be greater than `test_start`."
            _raise_error(msg)
        return train_start, train_end, test_start, test_end

    def _single_train_test_split(
        self, tag: Union[str, int], data: TimeSeriesData
    ) -> Tuple[Union[str, int], List[TrainTestData]]:

        time_index = data.time
        train_start_idx = np.searchsorted(time_index, self.train_start)
        train_end_idx = np.searchsorted(time_index, self.train_end, side="right")
        test_start_idx = np.searchsorted(time_index, self.test_start)
        test_end_idx = np.searchsorted(time_index, self.test_end, side="right")

        return tag, [
            TrainTestData(
                train=data[train_start_idx:train_end_idx],
                test=data[test_start_idx:test_end_idx],
            )
        ]


class RollingOriginDataParition(DataPartitionBase):
    """Class for rolling-origin data-partition method.

    A rolling-origin data-partition splits data over multiple iterations, wherein each iteration, the end of train set "slides" forward
    by a fixed amount. The size of the train data can either increase or keeps constant, and size of test data keeps constant.
    Moreover, we allow a gap between the train and test data sets. The purpose of the gap is to focus on the long range forecasting ability of the model.

    Attributes:
        start_train_frac: A float in (0, 1) for the initial fraction of data used for training.
        test_frac: A float in (0, 1) for the fraction of data used for testing. The test set is taken at sliding positions from `start_train_frac` up to the end of the dataset.
        window_frac: A float in [0, 1) for the fraction of window between train and test data, i.e., the start of test set is `window_frac` away from the end of train set. Default is 0 (i.e., no gap between train and test set).
        expanding_steps: Optional; An integer for the number of expanding steps (i.e., number of folds). Default is 1.
        constant_train_size: Optional; A boolean for whether training data size should be constant. Default is False.
        multi: Optional; A boolean representing whether using multi-processing or not. Default is True.
        max_core: Optional; An integer representing the number of cores for multiprocessing. Default is None, which sets `max_core = max((total_cores - 1) // 2, 1)`.

    Sample Usage:
        >>> stdp = RollingOriginDataParition(start_train_frac = 0.5, test_frac = 0.2, expanding_steps = 3)
        >>> res = stdp.split(data)
    """

    def __init__(
        self,
        start_train_frac: float,
        test_frac: float,
        window_frac: float = 0.0,
        expanding_steps: int = 1,
        constant_train_size: bool = False,
        multi: bool = True,
        max_core: Optional[int] = None,
    ) -> None:
        super(RollingOriginDataParition, self).__init__(multi=multi, max_core=max_core)

        start_train_frac, test_frac, window_frac, expanding_steps = self._param_check(
            start_train_frac, test_frac, window_frac, expanding_steps
        )
        self.start_train_frac = start_train_frac
        self.test_frac = test_frac
        self.window_frac = window_frac
        self.expanding_steps = expanding_steps
        self.constant_train_size = constant_train_size
        self.split_num = expanding_steps

    def _param_check(
        self,
        start_train_frac: float,
        test_frac: float,
        window_frac: float,
        expanding_steps: int,
    ) -> Tuple[float, float, float, int]:

        if start_train_frac < 0.0 or start_train_frac > 1.0:
            msg = f"`start_frain_frac` should be a float in (0, 1) but receives {start_train_frac}."
            _raise_error(msg)
        if test_frac < 0.0 or test_frac > 1.0:
            msg = f"`test_frac` should be a float in (0, 1) but receives {test_frac}."
            _raise_error(msg)
        if window_frac < 0.0 or window_frac > 1.0:
            msg = f"`window_frac` should be a float in [0, 1) but receives {window_frac}."
            _raise_error(msg)
        if start_train_frac + test_frac + window_frac > 1.0:
            msg = "The sum of `start_frain_frac` and `test_frac` should be no larger than 1."
            _raise_error(msg)
        if start_train_frac + test_frac + window_frac == 1 and expanding_steps > 1:
            msg = "When the sum of `start_frain_frac` and `test_frac` is 1, `expanding_steps` should be 1."
            _raise_error(msg)
            msg = "When the sum of `start_frain_frac` and `test_frac` is 1, `expanding_steps` should be 1."
            _raise_error(msg)
        if expanding_steps < 1:
            msg = f"`expanding_steps` should be a positive integer but receives {expanding_steps}."
        return start_train_frac, test_frac, window_frac, expanding_steps

    def _single_train_test_split(
        self, tag: Union[str, int], data: TimeSeriesData
    ) -> Tuple[Union[str, int], List[TrainTestData]]:
        n = len(data)
        start_train_size = _get_absolute_size(n, self.start_train_frac)
        test_size = _get_absolute_size(n, self.test_frac)
        window_size = _get_absolute_size(n, self.window_frac)
        offsets = np.linspace(
            0,
            n - start_train_size - test_size - window_size,
            self.expanding_steps,
            dtype=int,
        )
        if self.constant_train_size:  # does not increase training set length
            res = [
                TrainTestData(
                    train=data[offset : (offset + start_train_size)],
                    test=data[
                        (offset + start_train_size + window_size) : (
                            offset + start_train_size + test_size + window_size
                        )
                    ],
                )
                for offset in offsets
            ]
        else:
            res = [
                TrainTestData(
                    train=data[: (offset + start_train_size)],
                    test=data[
                        (offset + start_train_size + window_size) : (
                            offset + start_train_size + test_size + window_size
                        )
                    ],
                )
                for offset in offsets
            ]
        return tag, res



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Union
from typing import cast
from unittest import TestCase

import numpy as np
import pandas as pd
from parameterized import parameterized

from kats.consts import TimeSeriesData
from kats.data.utils import load_data
from kats.detectors.hourly_ratio_detection import HourlyRatioDetector


class HourlyRatioDectorTest(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        daily_data = load_data("peyton_manning.csv")
        daily_data.columns = ["time", "y"]
        cls.invalid_daily_ts = TimeSeriesData(daily_data)
        cls.invalid_minutely_ts = cls.data_generation(freq="T")

        cls.invalid_empty_ts = TimeSeriesData(pd.DataFrame([], columns=["time", "y"]))

        multi_data = load_data("multivariate_anomaly_simulated_data.csv")
        cls.invalid_multi_ts = TimeSeriesData(multi_data)

        cls.valid_hourly_ts = cls.data_generation(drop=False)
        cls.valid_missing_ts = cls.data_generation()
        cls.valid_minutely_ts = cls.data_generation(freq="T")

        plot_before_detector_ts = cls.data_generation(freq="T")
        cls.plot_before_detector_hr = HourlyRatioDetector(plot_before_detector_ts, freq="T", aggregate="max")

    @classmethod
    def data_generation(cls, freq: str = "H", drop: bool = True, frac: float = 0.95) -> TimeSeriesData:
        time = pd.date_range("2018-01-01", "2020-01-01", freq=freq)
        n = len(time)
        x = np.arange(n)
        values = np.abs(np.sin(30 * x)) + 0.3 * np.random.randn(n)
        df = pd.DataFrame(values, columns=["value"])
        df["time"] = time
        if drop:
            df = cast(pd.DataFrame, df.sample(frac=frac, replace=False))
        return TimeSeriesData(df)

    # pyre-ignore[16] Module parameterized.parameterized has no attribute expand.
    @parameterized.expand(
        [  # name, freq, aggregate
            ["valid_hourly_ts", "H", None],
            ["valid_missing_ts", "H", None],
            ["valid_minutely_ts", "T", "max"],
        ]
    )
    def test_detector(self, name: str, freq: str, aggregate: Union[str, None]) -> None:
        # test hourly data without missing vlaues
        ts = getattr(self, name)
        hr = HourlyRatioDetector(ts, freq=freq, aggregate=aggregate)
        hr._preprocess()
        hr.detector()
        hr.plot()

    # pyre-ignore[16] Module parameterized.parameterized has no attribute expand.
    @parameterized.expand(
        [
            ["invalid_daily_ts"],
            ["invalid_minutely_ts"],
            ["invalid_multi_ts"],
            ["invalid_empty_ts"],
        ]
    )
    def test_invalid(self, name: str) -> None:
        ts = getattr(self, name)
        self.assertRaises(ValueError, HourlyRatioDetector, ts)

    def test_minutely_other_aggregate(self) -> None:
        ts = self.invalid_minutely_ts
        self.assertRaises(ValueError, HourlyRatioDetector, data=ts, aggregate="other_method")

    def test_plot_before_detector(self) -> None:
        hr = self.plot_before_detector_hr
        self.assertRaises(ValueError, hr.plot)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


"""The Simple Heuristic model

Simple Heuristic model is a model that applies simple rules like mean or percentiles on historical data to get prediction.
"""
import logging
from typing import Any
from typing import Callable
from typing import Optional

import numpy as np
import pandas as pd

from kats.consts import Params
from kats.consts import TimeSeriesData
from kats.models.model import Model

MODELS = ["last", "mean", "median", "percentile"]


class SimpleHeuristicModelParams(Params):
    """Parameter class for Simple Heuristic model.

    Attributes:
        method: a method from MODELS that will be used on historical data
                "last" method will use the last values as prediction
                "mean" method will use the mean of historical data as prediction
                "median" method will use median of historical data as prediction
                "percentile" method will use percentile and alpha of historical data as prediction
        quantile: quantile to compute, which must be a number between 0 and 100 inclusive.
    """

    def __init__(self, method: str = "last", quantile: int = 95) -> None:
        super().__init__()
        self.method = method
        self.quantile = quantile
        logging.debug(f"Initialized SimpleHeuristicModel parameters with method: {self.method}")

    def validate_params(self) -> None:
        "Validate SimpleHeuristicModel Model Parameters"

        if self.method not in MODELS:
            msg = f"Only support {', '.join(MODELS)}"
            logging.error(msg)
            raise ValueError(msg)

        if not isinstance(self.quantile, int):
            msg = "Percentile needs to be a number from 0 to 100"
            logging.error(msg)
            raise ValueError(msg)

        if self.quantile < 0 or self.quantile > 100:
            msg = "Percentile needs to be a number from 0 to 100"
            logging.error(msg)
            raise ValueError(msg)


class SimpleHeuristicModel(Model[SimpleHeuristicModelParams]):
    """Model class for SimpleHeuristicModel Model.

    This class provides the fit, predict and plot methods for the Simple Heuristic Model

    Attributes:
        data: :class:`kats.consts.TimeSeriesData`, the input time series data as `TimeSeriesData`
        params: the parameter class defined with `SimpleHeuristicModelParams`
    """

    model: Callable[[np.ndarray], np.ndarray]
    include_history: bool = False
    dates: Optional[pd.DatetimeIndex] = None
    y_fcst: Optional[np.ndarray] = None
    y_fcst_lower: Optional[np.ndarray] = None
    y_fcst_upper: Optional[np.ndarray] = None
    fcst_df: pd.DataFrame = pd.DataFrame(data=None)
    freq: Optional[str] = None

    def __init__(self, data: TimeSeriesData, params: SimpleHeuristicModelParams) -> None:
        super().__init__(data, params)
        # pyre-fixme[16]: `Optional` has no attribute `value`.
        if not isinstance(self.data.value, pd.Series):
            msg = "Only support univariate time series, but get {type}.".format(type=type(self.data.value))
            logging.error(msg)
            raise ValueError(msg)

        self.model = self._calc_last
        self.y_fcst = None
        self.y_fcst_lower = None
        self.y_fcst_upper = None
        self.fcst_df = pd.DataFrame(data=None)

    def _calc_last(self, fitted_values: np.ndarray) -> np.ndarray:
        """subtracts last 'steps' data points from self.data.value

        Args:
            fitted_values: numpy ndarray with shape (len(self.data.value)//steps, steps) from self.data.value
        """
        return fitted_values[-1, :]

    def _calc_mean(self, fitted_values: np.ndarray) -> np.ndarray:
        """calculates mean from self.data.value
           returns ndarray with shape (steps,)

        Args:
            fitted_values: numpy ndarray with shape (len(self.data.value)//steps, steps) from self.data.value
        """
        return np.mean(fitted_values, 0)

    def _calc_median(self, fitted_values: np.ndarray) -> np.ndarray:
        """calculates median from self.data.value
           returns ndarray with shape (steps,)

        Args:
            fitted_values: numpy ndarray with shape (len(self.data.value)//steps, steps) from self.data.value
        """
        return np.median(fitted_values, 0)

    def _calc_percentile(self, fitted_values: np.ndarray) -> np.ndarray:
        """calculates percentile from self.data.value
           returns ndarray with shape (steps,)

        Args:
            fitted_values: numpy ndarray with shape (len(self.data.value)//steps, steps) from self.data.value
        """
        return np.percentile(fitted_values, self.params.quantile, 0)

    def fit(self) -> None:
        "fit Simple Heuristic Model."
        logging.debug("Call fit() with parameters: " "method:{method}".format(method=self.params.method))

        if self.params.method == "last":
            self.model = self._calc_last
        elif self.params.method == "mean":
            self.model = self._calc_mean
        elif self.params.method == "median":
            self.model = self._calc_median
        else:
            assert self.params.method == "percentile"
            self.model = self._calc_percentile

    # pyre-fixme[15]: `predict` overrides method defined in `Model` inconsistently.
    def predict(self, steps: int, *args: Any, include_history: bool = False, **kwargs: Any) -> pd.DataFrame:
        """predict with fitted Simple Heuristic Model.

        Args:
            steps: the steps or length of the prediction horizon
            include_history: whether to include the historical data in the prediction

        Returns:
            The predicted dataframe with the following columns:
                `time`, `fcst`, `fcst_lower`, and `fcst_upper`
        """
        logging.debug(f"Call predict() with parameters. Steps:{steps}")
        if self.model is None:
            raise ValueError("fit must be called before predict.")

        # TO DO: infer more robust frequency
        # pyre-fixme[16]: `Optional` has no attribute `time`.
        self.freq = kwargs.get("freq", pd.infer_freq(self.data.time))
        if self.freq is None:
            logging.info("Could not infer freq. Defaulting to daily 'D'")

        self.include_history = include_history

        # pyre-fixme[16]: `Optional` has no attribute `value`.
        fitted_len = (len(self.data.value) // steps) * steps
        fitted_values_s = self.data.value[-fitted_len:]
        fitted_values = np.reshape(np.asarray(fitted_values_s), [-1, steps])

        self.y_fcst = self.model(fitted_values)
        self.y_fcst_lower = self.y_fcst
        self.y_fcst_upper = self.y_fcst

        # create future dates
        last_date = self.data.time.max()
        dates = pd.date_range(start=last_date, periods=steps + 1, freq=self.freq)
        self.dates = dates[dates != last_date]

        if self.include_history:
            self.dates = np.concatenate((pd.to_datetime(self.data.time), self.dates))
            self.y_fcst = np.concatenate((np.asarray(self.data.value), self.y_fcst))
            self.y_fcst_lower = np.concatenate((np.asarray(self.data.value), self.y_fcst_lower))
            self.y_fcst_upper = np.concatenate((np.asarray(self.data.value), self.y_fcst_upper))

        self.fcst_df = pd.DataFrame(
            {
                "time": self.dates,
                "fcst": self.y_fcst,
                "fcst_lower": self.y_fcst_lower,
                "fcst_upper": self.y_fcst_upper,
            },
            copy=False,
        )
        logging.debug("Return forecast data: {fcst_df}".format(fcst_df=self.fcst_df))
        return self.fcst_df

    def __str__(self) -> str:
        return "Simple Heuristic Model"



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import inspect
import unittest
from typing import Any
from typing import Dict
from unittest.mock import ANY
from unittest.mock import patch

import pandas as pd

from kats.compat import compat
from kats.compat import pandas


class TestPandas(unittest.TestCase):
    assert_frame_equal_args: Dict[str, Any] = {}
    assert_series_equal_args: Dict[str, Any] = {}
    assert_index_equal_args: Dict[str, Any] = {}

    def setUp(self) -> None:
        for method in (
            "assert_frame_equal",
            "assert_series_equal",
            "assert_index_equal",
        ):
            args = {}
            setattr(self, f"{method}_args", args)
            for k in inspect.signature(getattr(pandas, method)).parameters:
                if k not in {"left", "right"}:
                    args[k] = ANY

    def test_version(self) -> None:
        self.assertTrue(pandas.version == compat.Version("pandas"))

    @patch("kats.compat.pandas.version", compat.Version("1.0"))
    @patch("kats.compat.pandas.pdt.assert_frame_equal")
    # pyre-fixme[2]: Parameter annotation cannot be `Any`.
    def test_assert_frame_equal_10(self, assert_frame_equal: Any) -> None:
        assert_frame_equal.return_value = False
        df = pd.DataFrame()
        result = pandas.assert_frame_equal(
            df, df, check_less_precise=2, check_flags=False, check_freq=False, rtol=0.01
        )
        self.assertFalse(result)
        args = dict(self.assert_frame_equal_args)
        args["check_less_precise"] = 2
        # drop args from other versions
        del args["rtol"]
        del args["atol"]
        del args["check_freq"]
        del args["check_flags"]
        assert_frame_equal.assert_called_once_with(df, df, **args)

    @patch("kats.compat.pandas.version", compat.Version("1.1"))
    @patch("kats.compat.pandas.pdt.assert_frame_equal")
    # pyre-fixme[2]: Parameter annotation cannot be `Any`.
    def test_assert_frame_equal_11(self, assert_frame_equal: Any) -> None:
        assert_frame_equal.return_value = False
        df = pd.DataFrame()
        result = pandas.assert_frame_equal(df, df, check_less_precise=2, check_flags=False, rtol=0.01)
        self.assertFalse(result)
        args = dict(self.assert_frame_equal_args)
        args["rtol"] = 0.01
        # drop args from other versions
        del args["check_less_precise"]
        del args["check_flags"]
        assert_frame_equal.assert_called_once_with(df, df, **args)

    @patch("kats.compat.pandas.version", compat.Version("1.2"))
    @patch("kats.compat.pandas.pdt.assert_frame_equal")
    # pyre-fixme[2]: Parameter annotation cannot be `Any`.
    def test_assert_frame_equal_12(self, assert_frame_equal: Any) -> None:
        assert_frame_equal.return_value = False
        df = pd.DataFrame()
        result = pandas.assert_frame_equal(df, df, check_less_precise=2, rtol=0.01)
        self.assertFalse(result)
        args = dict(self.assert_frame_equal_args)
        # drop args from other versions
        del args["check_less_precise"]
        args["rtol"] = 0.01
        assert_frame_equal.assert_called_once_with(df, df, **args)

    @patch("kats.compat.pandas.version", compat.Version("1.0"))
    @patch("kats.compat.pandas.pdt.assert_index_equal")
    @patch("kats.compat.pandas.pdt.assert_series_equal")
    def test_assert_series_equal_10(
        self,
        # pyre-fixme[2]: Parameter annotation cannot be `Any`.
        assert_series_equal: Any,
        # pyre-fixme[2]: Parameter annotation cannot be `Any`.
        assert_index_equal: Any,
    ) -> None:
        assert_series_equal.return_value = False
        s = pd.Series(dtype=int)
        result = pandas.assert_series_equal(s, s, check_less_precise=2, check_flags=False, check_freq=False, rtol=0.01)
        self.assertFalse(result)
        args = dict(self.assert_series_equal_args)
        args["check_less_precise"] = 2
        # drop args from other versions
        del args["check_category_order"]
        del args["check_freq"]
        del args["check_flags"]
        del args["rtol"]
        del args["atol"]
        del args["check_index"]
        assert_series_equal.assert_called_once_with(s, s, **args)

    @patch("kats.compat.pandas.version", compat.Version("1.0.2"))
    @patch("kats.compat.pandas.pdt.assert_index_equal")
    @patch("kats.compat.pandas.pdt.assert_series_equal")
    def test_assert_series_equal_102(
        self,
        # pyre-fixme[2]: Parameter annotation cannot be `Any`.
        assert_series_equal: Any,
        # pyre-fixme[2]: Parameter annotation cannot be `Any`.
        assert_index_equal: Any,
    ) -> None:
        # Check that on pandas 1.0.2, `check_less_precise` is passed and not later args
        assert_series_equal.return_value = False
        s = pd.Series(dtype=int)
        result = pandas.assert_series_equal(
            s,
            s,
            check_less_precise=2,
            check_category_order=True,
            check_flags=False,
            check_freq=False,
            rtol=0.01,
        )
        self.assertFalse(result)
        args = dict(self.assert_series_equal_args)
        args["check_less_precise"] = 2
        args["check_category_order"] = True
        # drop args from other versions
        del args["check_freq"]
        del args["check_flags"]
        del args["rtol"]
        del args["atol"]
        del args["check_index"]
        assert_series_equal.assert_called_once_with(s, s, **args)

    @patch("kats.compat.pandas.version", compat.Version("1.1"))
    @patch("kats.compat.pandas.pdt.assert_index_equal")
    @patch("kats.compat.pandas.pdt.assert_series_equal")
    def test_assert_series_equal_11(
        self,
        # pyre-fixme[2]: Parameter annotation cannot be `Any`.
        assert_series_equal: Any,
        # pyre-fixme[2]: Parameter annotation cannot be `Any`.
        assert_index_equal: Any,
    ) -> None:
        # Check that on pandas 1.1, `check_less_precise` is passed and not later args
        assert_series_equal.return_value = False
        s = pd.Series(dtype=int)
        result = pandas.assert_series_equal(
            s,
            s,
            check_less_precise=2,
            check_category_order=True,
            check_flags=False,
            check_freq=False,
            rtol=0.01,
        )
        self.assertFalse(result)
        args = dict(self.assert_series_equal_args)
        args["check_category_order"] = True
        args["check_freq"] = False
        args["rtol"] = 0.01
        # drop args from other versions
        del args["check_less_precise"]
        del args["check_flags"]
        del args["check_index"]
        assert_series_equal.assert_called_once_with(s, s, **args)

    @patch("kats.compat.pandas.version", compat.Version("1.2"))
    @patch("kats.compat.pandas.pdt.assert_index_equal")
    @patch("kats.compat.pandas.pdt.assert_series_equal")
    def test_assert_series_equal_12(
        self,
        # pyre-fixme[2]: Parameter annotation cannot be `Any`.
        assert_series_equal: Any,
        # pyre-fixme[2]: Parameter annotation cannot be `Any`.
        assert_index_equal: Any,
    ) -> None:
        # Check that on pandas 1.2, `check_less_precise` is not passed and not later args
        assert_series_equal.return_value = False
        s = pd.Series(dtype=int)
        result = pandas.assert_series_equal(
            s,
            s,
            check_less_precise=2,
            check_category_order=True,
            check_flags=False,
            check_freq=False,
            rtol=0.01,
        )
        self.assertFalse(result)
        args = dict(self.assert_series_equal_args)
        args["check_category_order"] = True
        args["check_freq"] = False
        args["check_flags"] = False
        args["rtol"] = 0.01
        # drop args from other versions
        del args["check_less_precise"]
        del args["check_index"]
        assert_series_equal.assert_called_once_with(s, s, **args)

    @patch("kats.compat.pandas.version", compat.Version("1.3"))
    @patch("kats.compat.pandas.pdt.assert_index_equal")
    @patch("kats.compat.pandas.pdt.assert_series_equal")
    def test_assert_series_equal_13(
        self,
        # pyre-fixme[2]: Parameter annotation cannot be `Any`.
        assert_series_equal: Any,
        # pyre-fixme[2]: Parameter annotation cannot be `Any`.
        assert_index_equal: Any,
    ) -> None:
        assert_series_equal.return_value = False
        s = pd.Series(dtype=int)
        result = pandas.assert_series_equal(
            s,
            s,
            check_less_precise=2,
            check_category_order=True,
            check_flags=False,
            check_freq=False,
            rtol=0.01,
            check_index=False,
        )
        self.assertFalse(result)
        args = dict(self.assert_series_equal_args)
        args["check_category_order"] = True
        args["check_freq"] = False
        args["check_flags"] = False
        args["rtol"] = 0.01
        args["check_index"] = False
        # drop args from other versions
        del args["check_less_precise"]
        assert_series_equal.assert_called_once_with(s, s, **args)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest
import unittest.mock as mock

from kats.compat import compat


class TestCompat(unittest.TestCase):
    def setUp(self) -> None:
        self.v = compat.Version("0.12.2")

    def test_Version_from_Version(self) -> None:
        result = compat.Version(self.v)
        self.assertEqual("0.12.2", result)

    def test_Version_from_packaging(self) -> None:
        result = compat.Version(self.v.version)
        self.assertEqual("0.12.2", result)

    @mock.patch("kats.compat.compat.metadata")
    def test_Version_module(self, metadata: object) -> None:
        metadata.version.return_value = "3.11.2"
        result = compat.Version("kats")
        self.assertEqual((3, 11, 2), result.version._key[1])

    @mock.patch("kats.compat.compat.metadata")
    def test_Version_module_truncated(self, metadata: object) -> None:
        metadata.version.return_value = "2.3.0"
        result = compat.Version("kats")
        self.assertEqual((2, 3), result.version._key[1])

    def test_Version_compare_str(self) -> None:
        self.assertTrue(self.v < "0.12.3")
        self.assertTrue(self.v <= "0.12.3")
        self.assertFalse(self.v > "0.12.2")
        self.assertTrue(self.v >= "0.12.2")
        self.assertTrue(self.v == "0.12.2")
        self.assertFalse(self.v == "0.12")
        self.assertTrue(self.v != "12.2")

    def test_Version_compare_Version(self) -> None:
        v12_3 = compat.Version("0.12.3")
        self.assertTrue(self.v < v12_3)
        self.assertTrue(self.v <= v12_3)
        self.assertFalse(self.v > self.v)
        self.assertTrue(self.v >= self.v)
        self.assertTrue(self.v == self.v)
        self.assertFalse(self.v == compat.Version("0.12"))
        self.assertTrue(self.v != compat.Version("12.2"))

    def test_incompatible_Version(self) -> None:
        def method(_x: compat.V, _y: compat.V) -> bool:
            raise TypeError()

        self.assertEqual(NotImplemented, self.v._compare(self.v, method))


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest
from typing import Dict
from unittest import TestCase

import numpy as np
import pandas as pd
import torch
from parameterized import parameterized

from kats.compat.pandas import assert_frame_equal
from kats.consts import TimeSeriesData
from kats.data.utils import load_air_passengers
from kats.data.utils import load_data
from kats.models.lstm import LSTMModel
from kats.models.lstm import LSTMParams
from kats.tests.models.test_models_dummy_data import (
    AIR_FCST_15_LSTM_PARAM_1_MODEL_1_MONTHLY,
)
from kats.tests.models.test_models_dummy_data import (
    AIR_FCST_15_LSTM_PARAM_1_MODEL_2_MONTHLY,
)
from kats.tests.models.test_models_dummy_data import (
    AIR_FCST_15_LSTM_PARAM_2_MODEL_1_MONTHLY,
)
from kats.tests.models.test_models_dummy_data import (
    AIR_FCST_15_LSTM_PARAM_2_MODEL_2_MONTHLY,
)
from kats.tests.models.test_models_dummy_data import (
    AIR_FCST_30_LSTM_PARAM_1_MODEL_1_MONTHLY,
)
from kats.tests.models.test_models_dummy_data import (
    AIR_FCST_30_LSTM_PARAM_1_MODEL_2_MONTHLY,
)
from kats.tests.models.test_models_dummy_data import (
    AIR_FCST_30_LSTM_PARAM_2_MODEL_1_MONTHLY,
)
from kats.tests.models.test_models_dummy_data import (
    AIR_FCST_30_LSTM_PARAM_2_MODEL_2_MONTHLY,
)
from kats.tests.models.test_models_dummy_data import (
    PT_FCST_15_LSTM_PARAM_1_MODEL_1_DAILY,
)
from kats.tests.models.test_models_dummy_data import (
    PT_FCST_15_LSTM_PARAM_1_MODEL_2_DAILY,
)
from kats.tests.models.test_models_dummy_data import (
    PT_FCST_15_LSTM_PARAM_2_MODEL_1_DAILY,
)
from kats.tests.models.test_models_dummy_data import (
    PT_FCST_15_LSTM_PARAM_2_MODEL_2_DAILY,
)
from kats.tests.models.test_models_dummy_data import (
    PT_FCST_30_LSTM_PARAM_1_MODEL_1_DAILY,
)
from kats.tests.models.test_models_dummy_data import (
    PT_FCST_30_LSTM_PARAM_1_MODEL_2_DAILY,
)
from kats.tests.models.test_models_dummy_data import (
    PT_FCST_30_LSTM_PARAM_2_MODEL_1_DAILY,
)
from kats.tests.models.test_models_dummy_data import (
    PT_FCST_30_LSTM_PARAM_2_MODEL_2_DAILY,
)

# pyre-fixme[5]: Global expression must be annotated.
AIR_TS = load_air_passengers()
PT_TS = TimeSeriesData(load_data("peyton_manning.csv"))
MULTI_DF = TimeSeriesData(load_data("multivariate_anomaly_simulated_data.csv"))
STEPS_1 = 15
STEPS_2 = 30
# pyre-fixme[5]: Global expression must be annotated.
TEST_DATA = {
    "monthly": {
        "ts": AIR_TS,
        "freq": "MS",
        "p1": LSTMParams(hidden_size=20, time_window=7, num_epochs=10),
        "p2": LSTMParams(hidden_size=10, time_window=3, num_epochs=4),
        "invalid_p": LSTMParams(hidden_size=-10, time_window=3, num_epochs=4),
        "invalid_ts": MULTI_DF,
        "m1": {
            "lr": 0.001,
        },
        "m2": {
            "lr": 0.0001,
        },
        "invalid_m1": {
            "lr": -0.001,
        },
        "steps_1": STEPS_1,
        "steps_2": STEPS_2,
        "truth_p1_m1_15": AIR_FCST_15_LSTM_PARAM_1_MODEL_1_MONTHLY,
        "truth_p2_m1_15": AIR_FCST_15_LSTM_PARAM_2_MODEL_1_MONTHLY,
        "truth_p1_m2_15": AIR_FCST_15_LSTM_PARAM_1_MODEL_2_MONTHLY,
        "truth_p2_m2_15": AIR_FCST_15_LSTM_PARAM_2_MODEL_2_MONTHLY,
        "truth_p1_m1_30": AIR_FCST_30_LSTM_PARAM_1_MODEL_1_MONTHLY,
        "truth_p2_m1_30": AIR_FCST_30_LSTM_PARAM_2_MODEL_1_MONTHLY,
        "truth_p1_m2_30": AIR_FCST_30_LSTM_PARAM_1_MODEL_2_MONTHLY,
        "truth_p2_m2_30": AIR_FCST_30_LSTM_PARAM_2_MODEL_2_MONTHLY,
    },
    "daily": {
        "ts": PT_TS,
        "freq": "D",
        "p1": LSTMParams(hidden_size=20, time_window=7, num_epochs=10),
        "p2": LSTMParams(hidden_size=10, time_window=3, num_epochs=4),
        "invalid_p": LSTMParams(hidden_size=-10, time_window=3, num_epochs=4),
        "m1": {
            "lr": 0.001,
        },
        "m2": {
            "lr": 0.0001,
        },
        "invalid_m1": {
            "lr": -0.001,
        },
        "steps_1": STEPS_1,
        "steps_2": STEPS_2,
        "truth_p1_m1_15": PT_FCST_15_LSTM_PARAM_1_MODEL_1_DAILY,
        "truth_p2_m1_15": PT_FCST_15_LSTM_PARAM_2_MODEL_1_DAILY,
        "truth_p1_m2_15": PT_FCST_15_LSTM_PARAM_1_MODEL_2_DAILY,
        "truth_p2_m2_15": PT_FCST_15_LSTM_PARAM_2_MODEL_2_DAILY,
        "truth_p1_m1_30": PT_FCST_30_LSTM_PARAM_1_MODEL_1_DAILY,
        "truth_p2_m1_30": PT_FCST_30_LSTM_PARAM_2_MODEL_1_DAILY,
        "truth_p1_m2_30": PT_FCST_30_LSTM_PARAM_1_MODEL_2_DAILY,
        "truth_p2_m2_30": PT_FCST_30_LSTM_PARAM_2_MODEL_2_DAILY,
    },
}


class LSTMModelTest(TestCase):
    # pyre-fixme[16]: Module `parameterized.parameterized` has no attribute `expand`.
    @parameterized.expand(
        [
            [
                "monthly_p1_m1",
                TEST_DATA["monthly"]["ts"],
                TEST_DATA["monthly"]["p1"],
                TEST_DATA["monthly"]["m1"],
                TEST_DATA["monthly"]["steps_1"],
                TEST_DATA["monthly"]["steps_2"],
                TEST_DATA["monthly"]["truth_p1_m1_15"],
                TEST_DATA["monthly"]["truth_p1_m1_30"],
                TEST_DATA["monthly"]["freq"],
            ],
            [
                "monthly_p2_m1",
                TEST_DATA["monthly"]["ts"],
                TEST_DATA["monthly"]["p2"],
                TEST_DATA["monthly"]["m1"],
                TEST_DATA["monthly"]["steps_1"],
                TEST_DATA["monthly"]["steps_2"],
                TEST_DATA["monthly"]["truth_p2_m1_15"],
                TEST_DATA["monthly"]["truth_p2_m1_30"],
                TEST_DATA["monthly"]["freq"],
            ],
            [
                "monthly_p1_m2",
                TEST_DATA["monthly"]["ts"],
                TEST_DATA["monthly"]["p1"],
                TEST_DATA["monthly"]["m2"],
                TEST_DATA["monthly"]["steps_1"],
                TEST_DATA["monthly"]["steps_2"],
                TEST_DATA["monthly"]["truth_p1_m2_15"],
                TEST_DATA["monthly"]["truth_p1_m2_30"],
                TEST_DATA["monthly"]["freq"],
            ],
            [
                "monthly_p2_m2",
                TEST_DATA["monthly"]["ts"],
                TEST_DATA["monthly"]["p2"],
                TEST_DATA["monthly"]["m2"],
                TEST_DATA["monthly"]["steps_1"],
                TEST_DATA["monthly"]["steps_2"],
                TEST_DATA["monthly"]["truth_p2_m2_15"],
                TEST_DATA["monthly"]["truth_p2_m2_30"],
                TEST_DATA["monthly"]["freq"],
            ],
            [
                "daily_p1_m1",
                TEST_DATA["daily"]["ts"],
                TEST_DATA["daily"]["p1"],
                TEST_DATA["daily"]["m1"],
                TEST_DATA["daily"]["steps_1"],
                TEST_DATA["daily"]["steps_2"],
                TEST_DATA["daily"]["truth_p1_m1_15"],
                TEST_DATA["daily"]["truth_p1_m1_30"],
                TEST_DATA["daily"]["freq"],
            ],
            [
                "daily_p2_m1",
                TEST_DATA["daily"]["ts"],
                TEST_DATA["daily"]["p2"],
                TEST_DATA["daily"]["m1"],
                TEST_DATA["daily"]["steps_1"],
                TEST_DATA["daily"]["steps_2"],
                TEST_DATA["daily"]["truth_p2_m1_15"],
                TEST_DATA["daily"]["truth_p2_m1_30"],
                TEST_DATA["daily"]["freq"],
            ],
            [
                "daily_p1_m2",
                TEST_DATA["daily"]["ts"],
                TEST_DATA["daily"]["p1"],
                TEST_DATA["daily"]["m2"],
                TEST_DATA["daily"]["steps_1"],
                TEST_DATA["daily"]["steps_2"],
                TEST_DATA["daily"]["truth_p1_m2_15"],
                TEST_DATA["daily"]["truth_p1_m2_30"],
                TEST_DATA["daily"]["freq"],
            ],
            [
                "daily_p2_m2",
                TEST_DATA["daily"]["ts"],
                TEST_DATA["daily"]["p2"],
                TEST_DATA["daily"]["m2"],
                TEST_DATA["daily"]["steps_1"],
                TEST_DATA["daily"]["steps_2"],
                TEST_DATA["daily"]["truth_p2_m2_15"],
                TEST_DATA["daily"]["truth_p2_m2_30"],
                TEST_DATA["daily"]["freq"],
            ],
        ]
    )
    def test_fcst(
        self,
        name: str,
        ts: TimeSeriesData,
        params: LSTMParams,
        model_params: Dict[str, int],
        steps_1: int,
        steps_2: int,
        truth_1: pd.DataFrame,
        truth_2: pd.DataFrame,
        freq: str,
    ) -> None:
        torch.manual_seed(0)
        m = LSTMModel(data=ts, params=params)
        m.fit(**model_params)
        res_1 = m.predict(
            steps=steps_1,
            freq=freq,
        ).reset_index(drop=True)
        res_2 = m.predict(
            steps=steps_2,
            freq=freq,
        ).reset_index(drop=True)
        assert_frame_equal(res_1, truth_1)
        assert_frame_equal(res_2, truth_2)

    # pyre-fixme[16]: Module `parameterized.parameterized` has no attribute `expand`.
    @parameterized.expand(
        [
            [
                "invalid_p",
                TEST_DATA["monthly"]["ts"],
                TEST_DATA["monthly"]["invalid_p"],
                TEST_DATA["monthly"]["m1"],
            ],
            [
                "invalid_m1",
                TEST_DATA["monthly"]["ts"],
                TEST_DATA["monthly"]["p2"],
                TEST_DATA["monthly"]["invalid_m1"],
            ],
            [
                "invalid_ts",
                TEST_DATA["monthly"]["invalid_ts"],
                TEST_DATA["monthly"]["p2"],
                TEST_DATA["monthly"]["m1"],
            ],
        ]
    )
    def test_invalid_params(
        self,
        name: str,
        ts: TimeSeriesData,
        params: LSTMParams,
        model_params: Dict[str, int],
    ) -> None:
        with self.assertRaises((ValueError, np.linalg.LinAlgError, NotImplementedError)):
            m = LSTMModel(data=ts, params=params)
            m.fit(**model_params)

    def test_exec_plot(self) -> None:
        m = LSTMModel(data=TEST_DATA["monthly"]["ts"], params=TEST_DATA["monthly"]["p2"])
        m.fit(**TEST_DATA["monthly"]["m1"])
        _ = m.predict(steps=STEPS_1)
        m.plot()

    def test_name(self) -> None:
        m = LSTMModel(data=TEST_DATA["monthly"]["ts"], params=TEST_DATA["monthly"]["p2"])
        self.assertEqual(m.__str__(), "LSTM")

    def test_search_space(self) -> None:
        m = LSTMModel(data=TEST_DATA["monthly"]["ts"], params=TEST_DATA["monthly"]["p2"])
        self.assertEqual(
            m.get_parameter_search_space(),
            [
                {
                    "name": "hidden_size",
                    "type": "choice",
                    "values": [
                        1,
                        11,
                        21,
                        31,
                        41,
                        51,
                        61,
                        71,
                        81,
                        91,
                        101,
                        111,
                        121,
                        131,
                        141,
                        151,
                        161,
                        171,
                        181,
                        191,
                        201,
                        211,
                        221,
                        231,
                        241,
                        251,
                        261,
                        271,
                        281,
                        291,
                        301,
                        311,
                        321,
                        331,
                        341,
                        351,
                        361,
                        371,
                        381,
                        391,
                        401,
                        411,
                        421,
                        431,
                        441,
                        451,
                        461,
                        471,
                        481,
                        491,
                    ],
                    "value_type": "int",
                    "is_ordered": True,
                },
                {
                    "name": "time_window",
                    "type": "choice",
                    "values": [
                        1,
                        2,
                        3,
                        4,
                        5,
                        6,
                        7,
                        8,
                        9,
                        10,
                        11,
                        12,
                        13,
                        14,
                        15,
                        16,
                        17,
                        18,
                        19,
                    ],
                    "value_type": "int",
                    "is_ordered": True,
                },
                {
                    "name": "num_epochs",
                    "type": "choice",
                    "values": [
                        50,
                        100,
                        150,
                        200,
                        250,
                        300,
                        350,
                        400,
                        450,
                        500,
                        550,
                        600,
                        650,
                        700,
                        750,
                        800,
                        850,
                        900,
                        950,
                        1000,
                        1050,
                        1100,
                        1150,
                        1200,
                        1250,
                        1300,
                        1350,
                        1400,
                        1450,
                        1500,
                        1550,
                        1600,
                        1650,
                        1700,
                        1750,
                        1800,
                        1850,
                        1900,
                        1950,
                    ],
                    "value_type": "int",
                    "is_ordered": True,
                },
            ],
        )

    # Filler test
    def test_validate_params(self) -> None:
        params = TEST_DATA["monthly"]["p2"]
        params.validate_params()


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
from typing import Any
from typing import Dict
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

import numpy as np
import pandas as pd
from statsmodels.tsa import holtwinters

from kats.compat import compat

ArrayLike = Union[np.ndarray, Sequence[float]]
Frequency = Union[int, str, pd.Timedelta]


version: compat.Version = compat.Version("statsmodels")


class HoltWintersResults(holtwinters.HoltWintersResults):
    smoothing_trend: float = 0.0

    def __init__(self, results: holtwinters.HoltWintersResults) -> None:
        if version < "0.12":
            # pyre-fixme[20]: Argument `aic` expected.
            super().__init__(results.model, results.params)
            self.params["smoothing_trend"] = self.params["smoothing_slope"]
        else:
            super().__init__(
                results.model,
                results.params,
                results.sse,
                results.aic,
                results.aicc,
                results.bic,
                results.optimized,
                results.level,
                results.trend,
                results.season,
                results.params_formatted,
                results.resid,
                results.k,
                results.fittedvalues,
                results.fittedfcast,
                results.fcastvalues,
                results.mle_retvals,
            )


class ExponentialSmoothing(holtwinters.ExponentialSmoothing):
    _use_boxcox: bool
    _initialization_method: str

    def __init__(
        self,
        endog: ArrayLike,
        trend: Optional[str] = None,
        damped_trend: bool = False,
        seasonal: Optional[str] = None,
        *,
        bounds: Optional[Dict[str, Tuple[float, float]]] = None,
        dates: Optional[Union[ArrayLike, pd.DatetimeIndex, pd.PeriodIndex]] = None,
        freq: Optional[Frequency] = None,
        initialization_method: str = "estimated",
        initial_level: Optional[float] = None,
        initial_seasonal: Optional[ArrayLike] = None,
        initial_trend: Optional[float] = None,
        missing: str = "none",
        seasonal_periods: Optional[int] = None,
        use_boxcox: bool = False,
    ) -> None:
        if version < "0.12":
            self._use_boxcox = use_boxcox
            self._initialization_method = initialization_method
            if bounds is not None:
                logging.warning("ExponentialSmoothing parameter 'bounds' not supported by statsmodels")
            if initialization_method != "missing":
                logging.warning("ExponentialSmoothing parameter 'initialization_method' not supported by statsmodels")
            if initial_level is not None:
                logging.warning("ExponentialSmoothing parameter 'initial_level' not supported by statsmodels")
            if initial_seasonal is not None:
                logging.warning("ExponentialSmoothing parameter 'initial_seasonal' not supported by statsmodels")
            if initial_trend is not None:
                logging.warning("ExponentialSmoothing parameter 'initial_trend' not supported by statsmodels")
            if seasonal_periods is not None:
                logging.warning("ExponentialSmoothing parameter 'seasonal_periods' not supported by statsmodels")
            super().__init__(
                endog,
                trend,
                damped_trend,
                seasonal,
                seasonal_periods,
                dates,
                freq,
                missing,
            )
        else:
            super().__init__(
                endog,
                trend,
                damped_trend,
                seasonal,
                seasonal_periods=seasonal_periods,
                initialization_method=initialization_method,
                initial_level=initial_level,
                initial_trend=initial_trend,
                initial_seasonal=initial_seasonal,
                use_boxcox=use_boxcox,
                bounds=bounds,
                dates=dates,
                freq=freq,
                missing=missing,
            )

    def fit(self, *args: Any, **kwargs: Any) -> holtwinters.HoltWintersResults:
        if version < "0.12":
            result = super().fit(*args, use_boxcox=self._use_boxcox, **kwargs)
        else:
            result = super().fit(*args, **kwargs)
        return HoltWintersResults(result)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# from typing import Dict
import unittest

from kats.compat import compat
from kats.compat import statsmodels


class TestStatsmodels(unittest.TestCase):
    def test_version(self) -> None:
        self.assertTrue(statsmodels.version == compat.Version("statsmodels"))


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
MetaDetectHptSelect is a meta learner that predict best hyper parameters of chosen detection algorithm given a time series features
before predicting, user needs to train the model:
    data_x=time_series features
    data_y=best (observed) hyper parameters for those features
"""

from __future__ import annotations

import logging
import math
from dataclasses import dataclass
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Set
from typing import Type
from typing import Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from kats.consts import TimeSeriesData
from kats.detectors.detector import DetectorModel
from kats.detectors.meta_learning.exceptions import KatsDetectorHPTIllegalHyperParameter
from kats.detectors.meta_learning.exceptions import (
    KatsDetectorHPTModelUsedBeforeTraining,
)
from kats.detectors.meta_learning.exceptions import KatsDetectorHPTTrainError
from kats.detectors.meta_learning.exceptions import KatsDetectorsUnimplemented
from kats.detectors.meta_learning.metalearning_detection_model import (
    PARAMS_TO_SCALE_DOWN,
)
from kats.detectors.meta_learning.metalearning_detection_model import (
    metadata_detect_preprocessor,
)
from kats.models.metalearner.metalearner_hpt import MetaLearnHPT
from kats.tsfeatures.tsfeatures import TsFeatures

_log: logging.Logger = logging.getLogger(__name__)


def get_ts_features(ts: TimeSeriesData) -> Dict[str, float]:
    """
    Extract TSFeatures for the input time series.
    """
    # Run Kats TsFeatures
    ts_features = TsFeatures(hw_params=False)
    feats = ts_features.transform(ts)

    # Rounding features
    features = {}
    assert isinstance(feats, dict)
    for feature_name, feature_val in feats.items():
        if not math.isnan(feature_val):
            feature_val = format(feature_val, ".4f")
        features[feature_name] = feature_val

    return features


def metadata_detect_reader(
    rawdata: pd.DataFrame,
    algorithm_name: str,
    params_to_scale_down: Set[str] = PARAMS_TO_SCALE_DOWN,
) -> Dict[str, pd.DataFrame]:
    """
    A helper function:
    preprocess meta data table to get the input for class MetaDetectHptSelect
    """
    # table's format: [{"hpt_res":..., "features":..., "best_model":...}, ...]
    table = metadata_detect_preprocessor(rawdata=rawdata, params_to_scale_down=params_to_scale_down)

    metadata = {}
    metadata["data_x"] = pd.DataFrame([item["features"] for item in table])
    metadata["data_y"] = pd.DataFrame([pd.Series(item["hpt_res"][algorithm_name][0]) for item in table])

    return metadata


@dataclass
class NNParams:
    """
    A dataclass that saves hyper-parameters for training neural networks

    scale: A boolean to specify whether or not to normalize time series features to zero mean and unit variance. Default is False.
    loss_scale: A float to specify the hyper-parameter to scale regression loss and classification loss,
        which controls the trade-off between the accuracy of regression task and classification task.
        A larger loss_scale value gives a more accurate prediction for classification part, and a lower value gives
        a more accurate prediction for regression part. Default is 1.0.
    lr: A float for learning rate. Default is 0.001.
    n_epochs: An integer for the number of epochs. Default is 1000.
    batch_size: An integer for the batch size. Default is 128.
    method: A string for the name of optimizer. Can be 'SGD' or 'Adam'. Default is 'SGD'.
    val_size: A float for the proportion of validation set of. It should be within (0, 1). Default is 0.1.
    momentum: A fload for the momentum for SGD. Default value is 0.9.
    n_epochs_stop:An integer or a float for early stopping condition. If the number of epochs is larger than n_epochs_stop
        and there is no improvement on validation set, we stop training. One can turn off the early stopping feature by
        setting n_epochs_stop = np.inf. Default is 20.

    """

    scale: bool = False
    loss_scale: float = 1.0
    lr: float = 0.001
    n_epochs: int = 1000
    batch_size: int = 128
    method: str = "SGD"
    val_size: float = 0.1
    momentum: float = 0.9
    n_epochs_stop: Union[int, float] = 20


class MetaDetectHptSelect:
    """A class for meta-learning framework on hyper-parameters tuning for detection algorithm.

    Attributes:
        data_x: A `pandas.DataFrame` object of time series features. data_x should not be None unless load_model is True. Default is None.
        data_y: A `pandas.DataFrame` object of the corresponding best hyper-parameters. data_y should not be None unless load_model is True. Default is None.
        detector_model: Type[DetectorModel]. A detector model in Kats.

    Sample Usage:
        >>> mdhs = MetaDetectHptSelect(data_x=datax, data_y=datay, detector_model=CUSUMDetectorModel)
        >>> mdhs.train(
                num_idx=["scan_window"],
                cat_idx=["threshold_low", "threshold_high"],
                n_hidden_shared=[20],
                n_hidden_cat_combo=[[5], [5]],
                n_hidden_num=[5],
                nnparams=self.nnparams,
            )  # train NNs
        >>> mdhs.plot()  # plot loss path
        >>> # generate prediction for future TSs
        >>> res = mdhs.get_hpt_from_features(np.random.normal(0, 1, [10, 40]))
    """

    def __init__(
        self,
        data_x: pd.DataFrame,
        data_y: pd.DataFrame,
        detector_model: Type[DetectorModel],
    ) -> None:
        self.detector_model: Type[DetectorModel] = detector_model
        self._check_valid_input(data_x, data_y, self.detector_model)
        self._data_x: pd.DataFrame = data_x
        self._data_y: pd.DataFrame = data_y
        self._algorithm_name: str = self.detector_model.__name__
        self._meta_hpt_model: Optional[MetaLearnHPT] = None

        # remove parameters which are constants
        # this causes the neural network training to fail otherwise
        self.const_params_dict: Dict[str, Any] = self.get_const_params(self._data_y)
        label_cols = [x for x in self._data_y.columns if x not in self.const_params_dict.keys()]
        self._data_y = self._data_y[label_cols]
        msg = """
            Removed parameters (columns) in data_y which are constants.
            Otherwise this causes the neural network training to fail.
        """
        _log.warning(msg=msg)

    @staticmethod
    def _check_valid_input(
        data_x: pd.DataFrame,
        data_y: pd.DataFrame,
        detector_model: Type[DetectorModel],
    ) -> None:
        # check: are all hyper parameters valid for the given algorithm
        for idx, hp in data_y.iterrows():
            try:
                # pyre-fixme[6]
                MetaDetectHptSelect._init_detection_model(detector_model, hp.to_dict())
            except Exception as e:
                raise KatsDetectorHPTIllegalHyperParameter(detector_model.__name__, idx, hp.to_dict(), e)

    @staticmethod
    def _init_detection_model(detector_model: Type[DetectorModel], hyper_parameters: Dict[str, Any]) -> None:
        """
        returns detection algorithm for given algorithm name initialized with given hyper parameters
        """
        hpt_to_remove = [
            "threshold_low",
            "threshold_high",
            "detection_window_sec",
            "fraction",
            "direction_1d",
        ]

        hpt_algo = {k: v for k, v in hyper_parameters.items() if k not in hpt_to_remove}

        # Try initializing the detector model
        # pyre-fixme[45]
        _ = detector_model(**hpt_algo)

    @staticmethod
    def get_const_params(df: pd.DataFrame) -> Dict[str, Any]:
        num_values_dict = df.nunique().to_dict()

        assert num_values_dict is not None and isinstance(num_values_dict, dict)
        const_params = [key for key in num_values_dict if num_values_dict[key] == 1]

        const_params_dict = {}
        for param in const_params:
            const_params_dict[param] = df[param].iloc[0]
        return const_params_dict

    def train(
        self,
        num_idx: List[str],
        cat_idx: List[str],
        n_hidden_shared: List[int],
        n_hidden_cat_combo: List[List[int]],
        n_hidden_num: List[int],
        nnparams: Optional[NNParams] = None,
    ) -> None:
        """Build a multi-task neural network and train it.

        This function builds a multi-task neural network according to given neural network structure
        (i.e., n_hidden_shared, n_hidden_cat_combo, n_hidden_num).

        Args:
            num_idx: A list of names of numerical parameters in the given data_y.
            cat_idx: A list of names of categorical parameters in the given data_y.
            n_hidden_shared: A list of numbers of hidden neurons in each shared hidden layer.
            n_hidden_cat_combo: A list of lists of task-specific hidden layers' sizes of each categorical response variables.
            n_hidden_num: A list of task-specific hidden layers' sizes of numerical response variables.
            nnparams: hyper-parameters used in training neural networks.
        """
        const_num_provided = set(num_idx) & set(self.const_params_dict.keys())
        const_cat_provided = set(cat_idx) & set(self.const_params_dict.keys())

        if len(const_num_provided) > 0:
            raise ValueError(f"{const_num_provided} is constant value in data_y, and we have removed that column.")

        if len(const_cat_provided) > 0:
            raise ValueError(f"{const_cat_provided} is constant value in data_y, and we have removed that column.")

        # n_hiddent cat_combo should have same dimensions as number
        # of categorical variables. Else, MetaLearnHPT throws an error
        # this is often difficult to set beforehand, since the data
        # may not choose different value for a param, during best HPT
        # calculation. Hence this check prevents the hpt training from
        # failing
        if len(n_hidden_cat_combo) != len(cat_idx):
            _log.info("Resetting HPT network to match cat dimensions")
            n_hidden_cat_combo = [[5]] * len(cat_idx)

        if nnparams is None:
            nnparams = NNParams()

        hpt_model = MetaLearnHPT(
            data_x=self._data_x,
            data_y=self._data_y,
            categorical_idx=cat_idx,
            numerical_idx=num_idx,
            scale=nnparams.scale,
        )

        hpt_model.build_network(
            n_hidden_shared=n_hidden_shared,
            n_hidden_cat_combo=n_hidden_cat_combo,
            n_hidden_num=n_hidden_num,
        )

        try:
            hpt_model.train(
                loss_scale=nnparams.loss_scale,
                lr=nnparams.lr,
                n_epochs=nnparams.n_epochs,
                batch_size=nnparams.batch_size,
                method=nnparams.method,
                val_size=nnparams.val_size,
                momentum=nnparams.momentum,
                n_epochs_stop=nnparams.n_epochs_stop,
            )
        except Exception as e:
            raise KatsDetectorHPTTrainError(e)

        self._meta_hpt_model = hpt_model

    def plot(self, **kwargs: Any) -> Sequence[plt.Axes]:
        model = self._meta_hpt_model
        if model is None:
            raise KatsDetectorHPTModelUsedBeforeTraining()
        return model.plot(**kwargs)

    def get_hpt_from_ts(self, ts: TimeSeriesData) -> pd.DataFrame:
        features_dict = get_ts_features(ts)
        features_array = np.asarray(features_dict.values())
        return self.get_hpt_from_features(features_array)

    def get_hpt_from_features(self, source_x: Union[np.ndarray, List[np.ndarray], pd.DataFrame]) -> pd.DataFrame:
        assert self._meta_hpt_model is not None

        ans = self._meta_hpt_model.pred_by_feature(source_x)
        ans_df = pd.DataFrame(ans)

        # add the const params that were not predicted by the neural net back
        for k, v in self.const_params_dict.items():
            ans_df[k] = v

        return ans_df

    def save_model(self) -> None:
        raise KatsDetectorsUnimplemented()

    def load_model(self) -> None:
        raise KatsDetectorsUnimplemented()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""Kats ensemble model

Implementation of the Kats ensemble model. It starts from seasonality detection, if seasonality detected, it
continues to perform STL decomposition, then fit forecasting models on de-seasonalized components and aggregate;
otherwise it simiply leverage individual forecasting models and ensembling. We provided two ensembling methods,
weighted average and median ensembling.
"""

from __future__ import annotations

import logging
import math
import multiprocessing
import sys
from copy import copy
from multiprocessing import cpu_count
from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Type
from typing import Union

import numpy as np
import pandas as pd

from kats.consts import Params
from kats.consts import TimeSeriesData

# Seasonality detector
from kats.detectors.seasonality import ACFDetector
from kats.models import arima
from kats.models import holtwinters
from kats.models import linear_model
from kats.models import quadratic_model
from kats.models import sarima
from kats.models import theta
from kats.models.ensemble.ensemble import EnsembleParams
from kats.models.model import Model
from kats.utils.backtesters import BackTesterSimple

# STL decomposition
from kats.utils.decomposition import TimeSeriesDecomposition

# from numpy.typing import ArrayLike
ArrayLike = Union[Sequence[float], np.ndarray]

# models that can fit de_seasonal component
MODELS = {
    "arima": arima.ARIMAModel,
    "holtwinters": holtwinters.HoltWintersModel,
    "sarima": sarima.SARIMAModel,
    "linear": linear_model.LinearModel,
    "quadratic": quadratic_model.QuadraticModel,
    "theta": theta.ThetaModel,
}

# models that can fit seasonal time series data
SMODELS = {
    "theta": theta.ThetaModel,
    # "sarima": sarima.SARIMAModel,
}


def _logged_error(msg: str) -> ValueError:
    """Log and raise an error."""
    logging.error(msg)
    return ValueError(msg)


# pyre-fixme[24]: Generic type `Model` expects 1 type parameter.
class KatsEnsemble(Model):
    """Decomposition based ensemble model in Kats
    This is the holistic ensembling class based on decomposition when seasonality presents
    """

    seasonality: bool = False
    sea_data: Optional[TimeSeriesData] = None
    desea_data: Optional[TimeSeriesData] = None
    steps: int = -1
    decomposition_method: str = ""
    model_params: Optional[EnsembleParams] = None
    fitted: Optional[Dict[str, Any]] = None
    weights: Optional[Dict[str, float]] = None
    predicted: Optional[Dict[str, pd.DataFrame]] = None
    err: Optional[Dict[str, float]] = None
    dates: Optional[pd.DatetimeIndex] = None
    fcst_dates: Optional[ArrayLike] = None
    fcst_df: Optional[pd.DataFrame] = None
    errors: Optional[Dict[str, Any]] = None

    def __init__(
        self,
        data: TimeSeriesData,
        params: Dict[str, Any],
    ) -> None:
        self.data = data
        self.freq: Optional[str] = pd.infer_freq(data.time)
        self.params = params
        self.validate_params()

    def validate_params(self) -> None:
        # validate aggregation method
        if self.params["aggregation"] not in ("median", "weightedavg"):
            method = self.params["aggregation"]
            msg = f"Only support `median` or `weightedavg` ensemble, but got {method}."
            raise _logged_error(msg)

        # validate decomposition method
        if self.params["decomposition_method"] in ("additive", "multiplicative"):
            self.decomposition_method = self.params["decomposition_method"]
        else:
            logging.info("Invalid decomposition method setting specified")
            logging.info("Defaulting to Additive Decomposition")
            self.decomposition_method = "additive"

        # validate m
        if (self.params["seasonality_length"] is not None) and (
            self.params["seasonality_length"] > int(len(self.data.time) // 2)
        ):
            msg = "seasonality_length value cannot be larger than"
            "1/2 of the length of give time series"
            raise _logged_error(msg)

        # check customized forecastExecutor
        if ("forecastExecutor" in self.params.keys()) and (self.params["forecastExecutor"] is not None):
            msg = "Using customized forecastExecutor from given parameters"
            logging.info(msg)
            self.forecastExecutor = self.params["forecastExecutor"]

        # check customized fitExecutor
        if ("fitExecutor" in self.params.keys()) and (self.params["fitExecutor"] is not None):
            msg = "Using customized fitExecutor from given parameters"
            logging.info(msg)
            self.fitExecutor = self.params["fitExecutor"]

    @staticmethod
    def seasonality_detector(data: TimeSeriesData) -> bool:
        """Detect seasonalities from given TimeSeriesData

        Args:
            data: :class:`kats.consts.TimeSeriesData`, the input `TimeSeriesData`

        Returns:
            Flag for the presence of seasonality
        """

        detector = ACFDetector(data)
        detector.detector()
        seasonality = detector.seasonality_detected
        return seasonality

    @staticmethod
    def deseasonalize(data: TimeSeriesData, decomposition_method: str) -> Tuple[TimeSeriesData, TimeSeriesData]:
        """STL decomposition to given TimeSeriesData

        Static method to perform decomposition on the input data

        Args:
            data: :class:`kats.consts.TimeSeriesData`, input time series data
            decomposition_method: the specific method for decomposition

        Returns:
            Tuple of seasonal data and de-seasonalized data
        """
        # create decomposer for time series decomposition
        decomposer = TimeSeriesDecomposition(data, decomposition_method)
        decomp = decomposer.decomposer()

        sea_data = copy(decomp["seasonal"])
        desea_data = copy(data)

        if decomposition_method == "additive":
            desea_data.value = desea_data.value - decomp["seasonal"].value
        else:
            desea_data.value = desea_data.value / decomp["seasonal"].value
        return sea_data, desea_data

    @staticmethod
    def reseasonalize(
        sea_data: TimeSeriesData,
        desea_predict: Dict[str, pd.DataFrame],
        decomposition_method: str,
        seasonality_length: int,
        steps: int,
    ) -> Dict[str, pd.DataFrame]:
        """Re-seasonalize the time series data

        Static method to re-seasonalize the input data

        Args:
            sea_data: :class:`kats.consts.TimeSeriesData`, the seasonal data from deseasonalize method
            desea_predict: dict of forecasted results for the deseasonalized
                data for each individual forecasting method
            decomposition_method: the specific method for decomposition
            seasonality_lenth: the length of seasonality
            steps: the length of forecasting horizon

        Returns:
            Dict of re-seasonalized data for each individual forecasting model
        """

        rep = math.trunc(1 + steps / seasonality_length)
        seasonality_unit = sea_data.value[-seasonality_length:]

        predicted = {}
        for model_name, desea_pred in desea_predict.items():
            if decomposition_method == "additive":
                if "fcst_lower" in desea_pred.columns and "fcst_upper" in desea_pred.columns:
                    # check consistency of time being index
                    if "time" in desea_pred.columns:
                        msg = "Setting time column as index"
                        logging.info(msg)
                        desea_pred.set_index("time", inplace=True)

                    # native C.I calculated from individual model
                    predicted[model_name] = (
                        desea_pred + np.tile(np.tile(seasonality_unit, rep)[:steps], [3, 1]).transpose()
                    )
                else:
                    # no C.I from individual model
                    tmp_fcst = desea_pred.fcst + np.tile(seasonality_unit, rep)[:steps]
                    predicted[model_name] = pd.DataFrame(
                        {
                            "time": desea_pred.index,
                            "fcst": tmp_fcst,
                            "fcst_lower": np.nan,
                            "fcst_upper": np.nan,
                        },
                        copy=False,
                    ).set_index("time")

            else:
                # multiplicative, element-wise multiply
                if "fcst_lower" in desea_pred.columns and "fcst_upper" in desea_pred.columns:
                    # check consistency of time being index
                    if "time" in desea_pred.columns:
                        msg = "Setting time column as index"
                        logging.info(msg)
                        desea_pred.set_index("time", inplace=True)

                    # native C.I calculated from individual model
                    predicted[model_name] = (
                        desea_pred * np.tile(np.tile(seasonality_unit, rep)[:steps], [3, 1]).transpose()
                    )
                else:
                    # no C.I from individual model
                    tmp_fcst = desea_pred.fcst * np.tile(seasonality_unit, rep)[:steps]
                    predicted[model_name] = pd.DataFrame(
                        {
                            "time": desea_pred.index,
                            "fcst": tmp_fcst,
                            "fcst_lower": 0,
                            "fcst_upper": 0,
                        },
                        copy=False,
                    ).set_index("time")

        return predicted

    def fitExecutor(
        self,
        data: TimeSeriesData,
        models: EnsembleParams,
        should_auto_backtest: bool = False,
        err_method: str = "mape",
    ) -> Tuple[Dict[str, Any], Optional[Dict[str, float]]]:
        """callable forecast executor

        This is native implementation with Python's multiprocessing
        fit individual model in `models` with given `data`. Services
        who use KatsEnsemble need to implement their own executor for better
        performance, if no executor function is given, the native version will be
        used.

        Attributes:
            data: :class:`kats.consts.TimeSeriesData`, given TimeSeriesData, could be original or de-seasonalized
            models: EnsembleParams object containing model params
                in BaseModelParams
            should_auto_backtest: boolean flag for additional back testing runs

        Returns:
            Tuple of fitted individual model and weights
        """

        # Fit individual model with given data
        num_process = min(len(MODELS), (cpu_count() - 1) // 2)
        if num_process < 1:
            num_process = 1
        pool = multiprocessing.Manager().Pool(processes=(num_process), maxtasksperchild=1000)

        fitted_models = {}
        for model in models.models:
            fitted_models[model.model_name] = pool.apply_async(
                self._fit_single,
                args=(
                    data,
                    MODELS[model.model_name.split("_")[0].lower()],
                    model.model_params,
                ),
            )
        pool.close()
        pool.join()
        fitted = {model: res.get() for model, res in fitted_models.items()}

        # if auto back testing
        weights = self.backTestExecutor(err_method) if should_auto_backtest else None
        return fitted, weights

    # pyre-fixme[15]: `fit` overrides method defined in `Model` inconsistently.
    def fit(self, err_method: str = "mape") -> KatsEnsemble:
        """Fit individual forecasting models via calling fitExecutor

        This is the fit methdo to fit individual forecasting model
        """

        self.seasonality = KatsEnsemble.seasonality_detector(self.data)

        # check if self.params["seasonality_length"] is given
        if self.seasonality and self.params["seasonality_length"] is None:
            msg = "The given time series contains seasonality,\
            a `seasonality_length` must be given in params."
            raise _logged_error(msg)

        # set up auto backtesting flag
        auto_backtesting = False if self.params["aggregation"] == "median" else True

        # check fitExecutor
        fitExecutor = self.params.get("fitExecutor")
        if fitExecutor is None:
            fitExecutor = self.fitExecutor

        if self.seasonality:
            # STL decomposition
            sea_data, desea_data = KatsEnsemble.deseasonalize(self.data, self.decomposition_method)
            self.sea_data = sea_data
            self.desea_data = desea_data

            # we created extra models
            given_models = copy(self.params["models"].models)
            for m in self.params["models"].models:
                if m.model_name.lower() in SMODELS.keys():
                    tmp = copy(m)
                    tmp.model_name = m.model_name + "_smodel"
                    given_models.append(tmp)

            self.model_params = model_params = EnsembleParams(given_models)
            self.fitted, self.weights = fitExecutor(
                data=desea_data,
                models=model_params,
                should_auto_backtest=auto_backtesting,
            )
        else:
            # fit models on the original data
            self.model_params = model_params = EnsembleParams(self.params["models"].models)
            self.fitted, self.weights = fitExecutor(
                data=self.data,
                models=model_params,
                should_auto_backtest=auto_backtesting,
                err_method=err_method,
            )
        return self

    # pyre-fixme[14]: `predict` overrides method defined in `Model` inconsistently.
    # pyre-fixme[15]: `predict` overrides method defined in `Model` inconsistently.
    def predict(self, steps: int) -> KatsEnsemble:
        """Predit future for each individual model

        Args:
            steps : number of steps ahead to forecast

        Returns:
            None
        """

        fitted = self.fitted
        if fitted is None:
            raise _logged_error("fit must be called before predict.")

        self.steps = steps
        if self.seasonality:
            sea_data = self.sea_data
            assert sea_data is not None
            # we should pred two types of model
            desea_fitted = {k: v for k, v in fitted.items() if "_smodel" not in k}
            desea_predict = {k: v.predict(self.steps).set_index("time") for k, v in desea_fitted.items()}

            # re-seasonalize
            predicted = KatsEnsemble.reseasonalize(
                sea_data=sea_data,
                desea_predict=desea_predict,
                decomposition_method=self.decomposition_method,
                seasonality_length=self.params["seasonality_length"],
                steps=self.steps,
            )

            # add extra model prediction results from smodels
            fitted_smodel = {k: v for k, v in fitted.items() if "_smodel" in k}
            extra_predict = {k: v.predict(self.steps).set_index("time") for k, v in fitted_smodel.items()}

            predicted.update(extra_predict)
            self.predicted = predicted
        else:
            predicted = {k: v.predict(self.steps).set_index("time") for k, v in fitted.items()}

            # add dummy C.I if the model doesn't have native C.I
            # this is a hack for median ensemble; everyone model needs to have
            # its native C.I if user choose weighted average ensemble.
            for k, v in predicted.items():
                # if predicted df doesn't have fcst_lower and fcst_upper
                if not {"fcst_lower", "fcst_upper"}.issubset(v.columns):
                    # add dummy C.I
                    tmp_v = copy(v)
                    tmp_v["fcst_lower"] = np.nan
                    tmp_v["fcst_upper"] = np.nan
                    predicted[k] = tmp_v
            self.predicted = predicted
        return self

    def forecast(self, steps: int) -> Tuple[Dict[str, pd.DataFrame], Optional[Dict[str, float]]]:
        """Holistic forecast method in Kats ensemble

        combine fit and predict methods to produce forecasted results
        this is especially useful for services which prefer to produce
        final forecasts without saving the fitted model

        Args:
            steps: the length of forecasting horizon

        Returns:
            Tuple of predicted values and weights
        """
        self.steps = steps
        self.seasonality = KatsEnsemble.seasonality_detector(self.data)

        # check if self.params["seasonality_length"] is given
        if (self.seasonality) and (self.params["seasonality_length"] is None):
            msg = "The given time series contains seasonality,\
            a `seasonality_length` must be given in params."
            raise _logged_error(msg)

        # set up auto backtesting flag
        auto_backtesting = False if self.params["aggregation"] == "median" else True

        if self.seasonality:
            # call forecastExecutor and move to next steps
            sea_data, desea_data = KatsEnsemble.deseasonalize(self.data, self.decomposition_method)
            self.sea_data = sea_data
            self.desea_data = desea_data

            # call forecasterExecutor with self.desea_data
            desea_predict, desea_err = self.forecastExecutor(
                data=desea_data,
                models=self.params["models"],
                steps=steps,
                should_auto_backtest=auto_backtesting,
            )
            # update the desea_predict with adding seasonality component
            # re-seasonalize
            predicted = KatsEnsemble.reseasonalize(
                sea_data=sea_data,
                desea_predict=desea_predict,
                decomposition_method=self.decomposition_method,
                seasonality_length=self.params["seasonality_length"],
                steps=self.steps,
            )

            # call forecasterExecutor with self.data
            # create new models
            # we created extra models
            extra_models = []
            for m in self.params["models"].models:
                if m.model_name.lower() in SMODELS.keys():
                    tmp = copy(m)
                    tmp.model_name = m.model_name + "_smodel"
                    extra_models.append(tmp)

            model_params = EnsembleParams(extra_models)
            extra_predict, extra_error = self.forecastExecutor(
                data=self.data,
                models=model_params,
                steps=self.steps,
                should_auto_backtest=auto_backtesting,
            )

            # combine with predict
            predicted.update(extra_predict)
            self.predicted = predicted

            if self.params["aggregation"] == "weightedavg":
                if desea_err is None:
                    desea_err = extra_error
                elif extra_error is not None:
                    desea_err.update(extra_error)
                self.err = forecast_error = desea_err
            else:
                forecast_error = None
        else:
            # no seasonality detected
            predicted, forecast_error = self.forecastExecutor(
                data=self.data,
                models=self.params["models"],
                steps=self.steps,
                should_auto_backtest=auto_backtesting,
            )
            self.err = forecast_error

            # same as in predict method above
            # add dummy C.I if the model doesn't have native C.I
            # this is a hack for median ensemble; everyone model needs to have
            # its native C.I if user choose weighted average ensemble.
            for k, v in predicted.items():
                # if predicted df doesn't have fcst_lower and fcst_upper
                if "fcst_lower" not in v.columns or "fcst_upper" not in v.columns:
                    # add dummy C.I
                    tmp_v = copy(v)
                    tmp_v["fcst_lower"] = np.nan
                    tmp_v["fcst_upper"] = np.nan
                    predicted[k] = tmp_v

            self.predicted = predicted

        # we need to transform err to weights if it's weighted avg
        if self.params["aggregation"] == "weightedavg":
            assert forecast_error is not None
            original_weights = {model: 1 / (err + sys.float_info.epsilon) for model, err in forecast_error.items()}
            self.weights = {model: err / sum(original_weights.values()) for model, err in original_weights.items()}
        else:
            self.weights = None
        return predicted, self.weights

    def forecastExecutor(
        self,
        data: TimeSeriesData,
        models: EnsembleParams,
        steps: int,
        should_auto_backtest: bool = False,
    ) -> Tuple[Dict[str, pd.DataFrame], Optional[Dict[str, float]]]:
        """Forecast Executor

        This is a callable execution function to
        (1). fit model
        (2). predict with a given steps
        (3). back testing (optional)

        Args:
            data: :class:`kats.consts.TimeSeriesData`, the input time series data as in :class:`kats.consts.TimeSeriesData`
            models: the ensemble parameters as in `EnsembleParams`
            steps: the length of forecasting horizon
            should_auto_backtest: flag to automatically perform back test, default as False

        Returns:
            The predicted values from each individual model and weights
        """

        # Fit individual model with given data
        num_process = min(len(MODELS), (cpu_count() - 1) // 2)
        if num_process < 1:
            num_process = 1
        pool = multiprocessing.Manager().Pool(processes=(num_process), maxtasksperchild=1000)

        fitted_models = {}
        for model in models.models:
            fitted_models[model.model_name] = pool.apply_async(
                self._fit_single,
                args=(
                    data,
                    MODELS[model.model_name.split("_")[0].lower()],
                    model.model_params,
                ),
            )
        pool.close()
        pool.join()
        fitted = {model: res.get() for model, res in fitted_models.items()}

        # simply predict with given steps
        predicted = {}
        for model_name, model_fitted in fitted.items():
            predicted[model_name] = model_fitted.predict(steps).set_index("time")

        # if auto back testing
        self.model_params = models  # used by _backtester_all
        if should_auto_backtest:
            _, errors = self._backtester_all()
        else:
            errors = None

        return predicted, errors

    def aggregate(self) -> pd.DataFrame:
        """Aggregate the results from predict method

        Args:
            None

        Returns:
            final results in pd.DataFrame
        """
        predicted = self.predicted
        if predicted is None:
            raise _logged_error("predict must be called before aggregate.")

        # create future dates
        last_date = self.data.time.max()
        dates = pd.date_range(start=last_date, periods=self.steps + 1, freq=self.freq)
        self.dates = dates = dates[dates != last_date]
        self.fcst_dates = dates.to_pydatetime()

        # collect the fcst, fcst_lower, and fcst_upper into dataframes
        fcsts = {}
        for col in ["fcst", "fcst_lower", "fcst_upper"]:
            fcsts[col] = pd.concat(
                [x[col].reset_index(drop=True) for x in predicted.values()],
                axis=1,
                copy=False,
            )
            fcsts[col].columns = predicted.keys()

        if self.params["aggregation"].lower() == "median":
            # clean up dataframes with C.I as np.nan or zero
            fcsts = self.clean_dummy_CI(fcsts, use_zero=False)
            self.fcst_df = fcst_df = pd.DataFrame(
                {
                    "time": dates,
                    "fcst": fcsts["fcst"].median(axis=1),
                    "fcst_lower": fcsts["fcst_lower"].median(axis=1),
                    "fcst_upper": fcsts["fcst_upper"].median(axis=1),
                },
                copy=False,
            )
        else:
            if fcsts["fcst_lower"].isnull().values.any() or fcsts["fcst_upper"].isnull().values.any():
                msg = "Conf. interval contains NaN, please check individual model."
                raise _logged_error(msg)
            weights = self.weights
            assert weights is not None
            weights = np.array(list(weights.values()))
            self.fcst_df = fcst_df = pd.DataFrame(
                {
                    "time": dates,
                    "fcst": fcsts["fcst"].dot(weights),
                    "fcst_lower": fcsts["fcst_lower"].dot(weights),
                    "fcst_upper": fcsts["fcst_upper"].dot(weights),
                },
                copy=False,
            )

        logging.debug("Return forecast data: {fcst_df}".format(fcst_df=fcst_df))
        return fcst_df

    @staticmethod
    def clean_dummy_CI(
        fcsts: Dict[str, pd.DataFrame],
        use_zero: bool = True,
    ) -> Dict[str, pd.DataFrame]:
        """Helper method to clean dummy prediction interval

        Args:
            fcsts: the dict of forecasting results from individual models
            use_zero: flag to use zero to fill nan, default as True

        Returns:
            the cleaned results in a dict
        """

        if use_zero:
            fcsts["fcst_lower"] = fcsts["fcst_lower"].fillna(0)
            fcsts["fcst_upper"] = fcsts["fcst_upper"].fillna(0)
        else:
            fcsts["fcst_lower"] = fcsts["fcst_lower"].replace(0, np.nan)
            fcsts["fcst_upper"] = fcsts["fcst_upper"].replace(0, np.nan)
        return fcsts

    def backTestExecutor(self, err_method: str) -> Dict[str, float]:
        """wrapper for back test executor

        services which use KatsEnsemble need to write their own backtest wrapper

        Args:
            None

        Returns:
            The dict of backtesting results
        """

        weights, _ = self._backtester_all(err_method=err_method)
        return weights

    def _fit_single(
        self,
        data: TimeSeriesData,
        # pyre-fixme[24]: Generic type `Callable` expects 2 type parameters.
        model_func: Callable,
        model_param: Params
        # pyre-fixme[24]: Generic type `Model` expects 1 type parameter.
    ) -> Model:
        """Private method to fit individual model

        Args:
            data: the input time series data
            model_func: the callable func to fit models
            model_param: the corresponding model parameter class

        Returns:
            Fitted Kats model
        """

        # get the model function call
        m = model_func(params=model_param, data=data)
        m.fit()
        return m

    def _backtester_single(
        self,
        params: Params,
        # pyre-fixme[24]: Generic type `Model` expects 1 type parameter.
        model_class: Type[Model],
        train_percentage: int = 80,
        test_percentage: int = 20,
        err_method: str = "mape",
    ) -> float:
        """Private method to run single back testing process

        Args:
            params: Kats model parameters
            model_class: Untyped. Defines type of model
            train_percentage: float. Percentage of data used for training
            test_percentage: float. Percentage of data used for testing
            error_method: list of strings indicating which errors to calculate
                we currently support "mape", "smape", "mae", "mase", "mse", "rmse"

        Returns:
            float, the backtesting error
        """

        bt = BackTesterSimple(
            [err_method],
            self.data,
            params,
            train_percentage,
            test_percentage,
            model_class,
        )
        bt.run_backtest()
        return bt.get_error_value(err_method)

    def _backtester_all(
        self,
        err_method: str = "mape",
    ) -> Tuple[Dict[str, float], Dict[str, float]]:
        """Private method to run all backtesting process

        Args:
            error_method: list of strings indicating which errors to calculate
                we currently support "mape", "smape", "mae", "mase", "mse", "rmse"

        Returns:
            Dict of errors from each model
        """
        model_params = self.model_params
        if model_params is None:
            raise _logged_error("fit must be called before backtesting.")

        num_process = min(len(MODELS.keys()), (cpu_count() - 1) // 2)
        if num_process < 1:
            num_process = 1
        pool = multiprocessing.Manager().Pool(processes=(num_process), maxtasksperchild=1000)
        backtesters = {}
        for model in model_params.models:
            backtesters[model.model_name] = pool.apply_async(
                self._backtester_single,
                args=(
                    model.model_params,
                    MODELS[model.model_name.split("_")[0].lower()],
                ),
                kwds={"err_method": err_method},
            )
        pool.close()
        pool.join()
        self.errors = errors = {model: res.get() for model, res in backtesters.items()}
        original_weights = {model: 1 / (err + sys.float_info.epsilon) for model, err in errors.items()}
        weights = {model: err / sum(original_weights.values()) for model, err in original_weights.items()}
        return weights, errors
